import torch
from torch import Tensor
from typing import List
from torch.linalg import solve, cholesky
from .augKernel import aug_kernel


def combined_kernel(x1, x2, decomposition, bws, scales):
    """This is the complete kernel k0 = sum_i ki
    Args:
        x1: A 'n1 x ndims' tensor
        x2: A 'n2 x ndims' tensor
        decomposition: list
        bws: A 'numGroups' tensor
        scales: A 'numGroups' tensor
    :Returns
        k: A 'n1 x n2' tensor
    """
    num_groups = len(decomposition)
    n1, n2 = x1.shape[0], x2.shape[0]
    k = torch.zeros(n1, n2, dtype=x1.dtype)
    for i in range(num_groups):
        coord = decomposition[i]
        bw = bws[i]
        scale = scales[i]
        k += aug_kernel(x1, x2, coord, bw, scale)
    return k


def combined_kernel_noise(
    x: Tensor,
    decomposition: List[Tensor],
    bws: Tensor,
    scales: Tensor,
    noises: Tensor,
    common_noise: float,
):
    return combined_kernel(x, x, decomposition, bws, scales) + (
        torch.sum(noises) + common_noise
    ) * torch.eye(x.shape[0], dtype=x.dtype)


def combined_kernel_mus(
    x1: Tensor, x2: Tensor, decomposition: List[Tensor], bws: Tensor, scales: Tensor
):
    """This is the complete kernel k0 = sum_i ki
    Args:
        x1: A 'n1 x ndims' tensor
        x2: A 'n2 x ndims' tensor
        decomposition: list
        bws: A 'numGroups' tensor
        scales: A 'numGroups' tensor
    :Returns
        k: A 'n1 x n2' tensor
    """
    num_groups = len(decomposition)
    n1, n2 = x1.shape[0], x2.shape[0]
    k = torch.zeros(n1, n2).to(x1)
    mus = torch.zeros(num_groups, dtype=torch.double)
    for i in range(num_groups):
        coord = decomposition[i]
        bw = bws[i]
        scale = scales[i]
        k11 = aug_kernel(x1, x2, coord, bw, scale)
        k += k11
        ell = cholesky(k11 + 1e-10 * torch.eye(k11.shape[0]).to(k11))
        temp = solve(ell, torch.ones(ell.shape[0], 1).to(ell))
        mus[i] = torch.squeeze(temp.T @ temp)
    return k, mus


def combined_kernel_noise_mus(
    x: Tensor,
    decomposition: List[Tensor],
    bws: Tensor,
    scales: Tensor,
    noises: Tensor,
    common_noise: float,
):
    kxx, mus = combined_kernel_mus(x, x, decomposition, bws, scales)
    return kxx + (torch.sum(noises) + common_noise) * torch.eye(x.shape[0]), mus



from .alebo import alebo
from .rembo import rembo
from .saasbo import saasbo
from .turbo import turbo
from .add_bo_quasi_nt.main import add_gp_bo

__all__ = ["alebo", "rembo", "saasbo", "turbo", "add_gp_bo"]



from tqdm import tqdm
import torch
import numpy as np


def random_optimize(func, bounds, n_iterations, random_kwargs=None):
    lowB, upB = bounds[:, 0], bounds[:, 1]

    pbar = tqdm(total=n_iterations)
    lb, ub = torch.tensor(lowB), torch.tensor(upB)
    X = torch.rand(n_iterations, func.dim)
    X = X * (ub - lb) + lb
    Y = list()
    for i in range(n_iterations):
        y = func(X[i])
        Y.append(y)
        pbar.update()
    if func.if_torch:
        Y = torch.stack(Y, dim=0)
    else:
        Y = np.stack(Y, axis=0)

    X = X.cpu().numpy()
    if isinstance(Y, torch.Tensor):
        Y = Y.cpu().numpy()
    best_x, best_y = X[np.argmax(Y)], np.max(Y)
    result = {"best_x": best_x, "best_y": best_y, "x_history": X, "y_history": Y}
    return best_x, best_y, result



import numpy as np
import torch
from .utils.logging import logger


def get_func_y_bound(args, method, n_repeats=5):
    min_value = np.inf
    max_value = -np.inf
    # optimize for min value
    name = args.func.name
    dim = args.func.dim
    logger.info(f"func:{name}, dim:{dim}, optimizer:{method.__name__}")
    args.func.switch_maximum()  # for optimizers for maximum
    method_kwargs = dict()
    if "method_kwargs" in dir(args):
        for method_name in ["botorch", "hebo", "turbo", "rembo", "saasbo", "alebo"]:
            if (method_name in method.__name__) and (
                method_name in args.method_kwargs.keys()
            ):
                method_kwargs = args.method_kwargs[method_name]
                break

    for i_repeats in range(n_repeats):
        _, best_y, _ = method(
            args.func, args.func.bound, args.numIterations, method_kwargs
        )
        if isinstance(best_y, torch.Tensor):
            best_y = best_y.detach().cpu().numpy()
        min_value = min(min_value, best_y)
        logger.info("min_value at iter %d: %.4f." % (i_repeats, min_value))

    # optimize for max value
    args.func.switch_minium()
    for i_repeats in range(n_repeats):
        _, best_y, _ = method(
            args.func, args.func.bound, args.numIterations, method_kwargs
        )
        if isinstance(best_y, torch.Tensor):
            best_y = best_y.detach().cpu().numpy()
        max_value = max(max_value, -best_y)
        logger.info("max_value at iter %d: %.4f." % (i_repeats, max_value))

    return min_value, max_value



def combined_mean_func(x, common_mean_func, mean_funcs, decomposition):
    """
    The total mean function (obtained by adding the common and individual mean function
    :param x:
    :param common_mean_func:
    :param mean_funcs:
    :param decomposition: list with elements
    :return mu0: (num,)
    """
    mu0 = common_mean_func(x)
    num_groups = len(decomposition)
    for k in range(num_groups):
        coord = decomposition[k]
        mu0 += mean_funcs[k](x[:, coord])
    return mu0



from math import log2
from tqdm import tqdm
import torch
from botorch.models.gp_regression import SingleTaskGP
from botorch.fit import fit_gpytorch_mll
from botorch.utils import standardize
import numpy as np
from scipy.stats.qmc import Sobol
from gpytorch.mlls import ExactMarginalLogLikelihood
from botorch.acquisition import UpperConfidenceBound
from botorch.optim import optimize_acqf


def suggest(input_x, input_y, bounds, n_suggestion=1, device="cpu"):
    # Standardize to 0-mean, 1-variance.
    train_y = standardize(input_y)  # shape(num, 1)
    min_x, max_x = torch.min(input_x, dim=0).values, torch.max(input_x, dim=0).values

    # min-max scale
    train_x = (input_x - min_x) / (max_x - min_x)
    scale_bounds = bounds.T
    scale_bounds = (scale_bounds - min_x) / (max_x - min_x)

    # 2. fit a model
    gp = SingleTaskGP(train_x, train_y)

    mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
    fit_gpytorch_mll(mll)

    # 3. Construct an acquisition function:
    UCB = UpperConfidenceBound(gp, beta=0.1)

    # 4. Optimize the acquisition function:
    candidate, acq_value = optimize_acqf(
        UCB,
        bounds=scale_bounds,
        q=n_suggestion,
        num_restarts=64,
        raw_samples=64,
    )
    # invert min-max scale
    candidate = candidate * (max_x - min_x) + min_x

    return candidate


def botorch_optimize(func, bounds, n_iterations, method_kwargs=dict()):
    """
    Maximize
    """
    ndims = bounds.shape[0]
    device = method_kwargs["device"] if "device" in method_kwargs.keys() else "cpu"
    num_rnd = (
        method_kwargs["n_init"] if "n_init" in method_kwargs.keys() else 10
    )  # num of random samples

    if not isinstance(bounds, torch.Tensor):
        bounds = torch.tensor(bounds, dtype=torch.float32, device=device)
    # randomly sample in the bounds.
    sampler = Sobol(d=ndims, scramble=True)
    if num_rnd & (num_rnd - 1) == 0:  # There exists integer m such that n=2^m.
        m = int(log2(num_rnd))
        input_x = sampler.random_base2(m)
    else:
        input_x = sampler.random(num_rnd)
    input_x = torch.tensor(input_x, dtype=torch.float32, device=device)
    input_x = (
        input_x * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]
    )  # shape(num_rnd, ndims)

    # Evaluate the value
    output_y = func(input_x)  # shape(num,)
    output_y = torch.tensor(output_y, dtype=torch.float32, device=device)

    # Record the maximal
    history = torch.zeros(n_iterations)
    max_idx = 0
    maxVal = output_y[0]
    history[0] = maxVal

    pbar = tqdm(total=n_iterations)
    pbar.update()
    for i in range(1, num_rnd):
        if output_y[i] > maxVal:
            maxVal = output_y[i]
            max_idx = i
        history[i] = maxVal
        pbar.update()

    for i in np.arange(num_rnd, n_iterations):
        candidate = suggest(
            input_x=input_x,
            input_y=output_y.unsqueeze(-1),
            bounds=bounds,
            n_suggestion=1,
            device=device,
        )
        # evaluate the value of candidate
        value = func(candidate)
        value = torch.tensor(value, dtype=torch.float32, device=device)
        # add them to input and output set.
        input_x = torch.vstack((input_x, candidate))  # shape(num+1, ndims)
        output_y = torch.hstack((output_y, value))  # shape(num+1,)
        # record the maximal
        if value > maxVal:
            maxVal = value
            max_idx = i
        history[i] = maxVal
        pbar.update(1)

    while maxVal.ndim > 0:
        maxVal = maxVal[0]
    maxPt = input_x[max_idx]
    result = {
        "best_x": maxPt,
        "best_y": maxVal,
        "x_history": input_x,
        "y_history": output_y,
        "best_y_history": history,
    }
    return maxPt, maxVal, result



import numpy as np
import torch

from .FunctionBase import TestFunction


class ProductSines(TestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(dim, "ProductSines")

        self._x_bound = np.array(
            [[-np.pi / 2, np.pi / 2]] * self.dim, dtype=float
        )  # shape(dim, 2)
        self.y_scale = 2
        self._min_pt = np.ones((dim,)) * np.pi / 2
        self._min_val = self._evaluate(self._min_pt)

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        prod_f = -1
        for i in range(self.dim):
            if isinstance(x, torch.Tensor):
                prod_f *= torch.sin(x[..., i])
            else:
                prod_f *= np.sin(x[..., i])
        return prod_f



# Define a function to be optimized.
# Here we use a simple synthetic function with a d=2 true linear embedding, in
# a D=100 ambient space.

import numpy as np

# pip install ax-platform
from ax.modelbridge.strategies.alebo import ALEBOStrategy


def alebo(eval_func, D, d, n_init, total_trials, **kwargs):

    # Define the parameters in the format expected by Ax.
    parameters = [
        {"name": f"x{i}", "type": "range", "bounds": [0.0, 1.0], "value_type": "float"}
        for i in range(D)
    ]
    # Setup the ALEBO optimization strategy
    alebo_strategy = ALEBOStrategy(D=D, d=d, init_size=n_init)

    # Run the optimization loop with that strategy
    # This will take about 30 mins to run

    from ax.service.managed_loop import optimize

    best_parameters, values, experiment, model = optimize(
        parameters=parameters,
        objective_name="objective",
        evaluation_function=eval_func,
        total_trials=total_trials,
        generation_strategy=alebo_strategy,
    )
    Y = np.array([trial.objective_mean for trial in experiment.trials.values()])
    X = np.array(
        [
            np.array([trial.arm.parameters.get(f"x{i}") for i in range(D)])
            for trial in experiment.trials.values()
        ]
    )
    return X, Y


if __name__ == "__main__":
    from ax.utils.measurement.synthetic_functions import branin
    import numpy as np

    DIM = 100
    EM_DIM = 4

    def branin_evaluation_function(parameterization):
        # Evaluates Branin on the first two parameters of the parameterization.
        # Other parameters are unused.
        lb, ub = np.full((DIM,), -5), np.full((DIM,), 10)
        lb[1], ub[1] = 0, 15
        x = np.array(
            [parameterization.get(f"x{i}") for i in range(len(parameterization))]
        )
        x = lb + (ub - lb) * x
        return {"objective": (-1 * branin(x[..., :2]), 0.0)}

    X, Y = alebo(branin_evaluation_function, D=DIM, d=EM_DIM, n_init=5, total_trials=30)
    Y_np = -1 * Y
    from matplotlib import pyplot as plt

    fig = plt.figure(figsize=(12, 6))
    ax = fig.add_subplot(111)
    ax.grid(alpha=0.2)
    ax.plot(range(1, 31), np.minimum.accumulate(Y_np))
    ax.axhline(y=branin.fmin, ls="--", c="k")
    ax.set_xlabel("Iteration")
    ax.set_ylabel("Best objective found")
    plt.savefig("results.png")



from .random import random_optimize
from .botorch_wrapper import botorch_optimize
from .hebo_wrapper import hebo_optimize
from .addbo_wrapper import addbo_optimize
from .rembo_wrapper import rembo_optimize
from .alebo_wrapper import alebo_optimize
from .turbo_wrapper import turbo_optimize
from .saasbo_wrapper import saasbo_optimize


__all__ = [
    "random_optimize",
    "botorch_optimize",
    "hebo_optimize",
    "rembo_optimize",
    "saasbo_optimize",
    "turbo_optimize",
    "alebo_optimize",
    "addbo_optimize",
]



import numpy as np


def get_hungarian_cost_matrix(a, cost_idx, power=2):
    """
    The objective is to take A - an orthogonal matrix and return the 'closest' permutation matrix.
    We can treat this as an assignment problem where the cost of assigning the Ai to Zk
    (where Z is the output Perm matrix) is as follows:
    cost_idx = 1, Cik = 1 - |Aik|,
    cost_idx = 2, Cik = 1 - |Aik| + \sum_{j=/=i} |Ajk|
    cost_idx = 3, Cik = 1 - |Aik| + \sum_{j=/=i} |Ajk| + \sum_{l=/=k}|Ail|
    :param a:
    :param cost_idx:
    :param power:
    :return:
    """
    # Prelims
    d, p = a.shape
    a_abs = np.abs(a) ** power
    c = 1 - a_abs
    if cost_idx >= 2:
        for i in range(d):
            c1 = np.sum(np.vstack((a_abs[1:i], a_abs[i + 1 :])), axis=0)
            c[i] = c[i] + c1
        if cost_idx == 3:
            for j in range(p):
                c2 = np.sum(np.hstack((a_abs[:, 1:j], a_abs[:, j + 1 :])), axis=1)
                c[:, j] = c[:, j] + c2
    return c



"""
git clone https://github.com/automl/HPOBench.git
cd HPOBench
pip install .
# for xgboost
pip install .[xgboost]
"""

import numpy as np
import importlib
from HDBOBenchmark.funcs.base.FunctionBase import TestFunction


class HPOBench(TestFunction):
    def __init__(
        self,
        benchname="ml.XGBoostBenchmark",
        task_id=167149,
        dim=None,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
        seed=1,
    ):
        packagename, benchname = benchname.split(".")
        Benchmark = importlib.import_module(
            "hpobench.benchmarks.%s" % packagename
        ).__getattribute__(benchname)

        self.b = Benchmark(task_id=task_id)
        self.config = self.b.get_configuration_space(seed=seed)
        self.seed = seed

        super().__init__(dim=len(self.config), name=benchname)

        self._x_bound = np.array([[-1, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        for i, variable in enumerate(self.config.items()):
            variable = variable[1]
            if "Float," in str(variable):
                self._x_bound[i] = (variable.lower, variable.upper)
            elif "Integer," in str(variable):
                self._x_bound[i] = (variable.lower, variable.upper + 1)
            else:
                raise NotImplementedError
        y_scale_dict = {}
        min_val_dict = {}
        self.y_scale = (
            y_scale_dict[benchname] if benchname in y_scale_dict.keys() else 100
        )
        self._min_pt = None
        self._min_val = (
            min_val_dict[benchname] if benchname in min_val_dict.keys() else 0
        )
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        if x.ndim == 1:
            return self._evaluate_hpobench(x)
        else:
            x_shape = x.shape
            x = x.reshape(-1, x.shape[-1])
            result = list()
            for x_slice in x:
                result.append(self._evaluate_hpobench(x_slice))
            result = np.array(result)
            result = result.reshape(x_shape[:-1])
            return result

    def _evaluate_hpobench(self, x) -> float:
        assert len(x) >= self.dim
        for i, name in enumerate(self.config.keys()):
            if i >= self.dim:
                break
            v = x[i]
            if "Integer," in str(self.config[name]):
                v = int(v)
            v = np.clip(v, self.config[name].lower, self.config[name].upper)
            self.config[name].default_value = v
            assert self.config[name].is_legal(
                self.config[name].default_value
            ), f"{self.config[name].default_value} is not legal for {str(self.config[name])}."

        result_dict = self.b.objective_function(
            configuration=self.config.get_default_configuration(),
            fidelity=None,
            rng=self.seed,
        )
        return result_dict["function_value"]



import torch
from torch import Tensor
from typing import Callable, Tuple
import numpy as np
from scipy.optimize import minimize, Bounds
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_optim.parameter_constraints import (
    make_scipy_bounds,
    _arrayify,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_optim.utils import columnwise_clamp


def gen_candidates_scipy(
    initial_conditions: Tensor,
    acquisition_function: Callable,
    lower_bounds: Tensor,
    upper_bounds: Tensor,
) -> Tuple[Tensor, Tensor]:
    """Generate a set of candidates using 'scipy.optimize.minimize'.

    Optimizes an acquisition function starting from a set of initial candidates
    using 'scipy.optimize.minimize' via a numpy converter.

    Args:
        initial_conditions: Starting points for optimization.
        acquisition_function: Acquisition function to be used.
        lower_bounds: Minimum values for each column of initial_conditions.
        upper_bounds: Maximum values for each column of initial_conditions.

    Returns:
        2-element tuple containing

        - The set of generated candidates.
        - The acquisition value for each t-batch.
    """
    ndims = initial_conditions.shape[-1]
    clamped_candidates = columnwise_clamp(
        X=initial_conditions, lower=lower_bounds, upper=upper_bounds
    )  # shape(num_restarts, q, ndims)

    shapeX = clamped_candidates.shape
    x0 = clamped_candidates.view(-1)
    bounds = make_scipy_bounds(
        X=initial_conditions, lower_bounds=lower_bounds, upper_bounds=upper_bounds
    )

    def f_np_wrapper(x: np.ndarray, f: Callable):
        """Given a torch callable, compute value + grad given a numpy array."""
        if np.isnan(x).any():
            raise RuntimeError(
                f"{np.isnan(x).sum()} elements of the {x.size} element array "
                f"`x` are NaN."
            )
        X = (
            torch.from_numpy(x)
            .to(initial_conditions)
            .view(-1, ndims)  # shape(num_restarts x q, ndims)
            .contiguous()
            .requires_grad_(True)
        )
        loss = f(X).sum()
        # compute gradient w.r.t. the inputs (does not accumulate in leaves)
        gradf = _arrayify(torch.autograd.grad(loss, X)[0].contiguous().view(-1))
        if np.isnan(gradf).any():
            msg = (
                f"{np.isnan(gradf).sum()} elements of the {x.size} element "
                "gradient array `gradf` are NaN. This often indicates numerical issues."
            )
            if initial_conditions.dtype != torch.double:
                msg += " Consider using `dtype=torch.double`."
            raise RuntimeError(msg)
        fval = loss.item()
        return fval, gradf

    x0 = _arrayify(x0)

    def f(x):
        return -acquisition_function(x)

    # To maximize the acquisition_function.
    res = minimize(
        fun=f_np_wrapper, args=(f,), x0=x0, method="L-BFGS-B", jac=True, bounds=bounds
    )
    candidates = torch.from_numpy(res.x).to(initial_conditions).reshape(shapeX)

    clamped_candidates = columnwise_clamp(
        X=candidates, lower=lower_bounds, upper=upper_bounds, raise_on_violation=True
    )
    with torch.no_grad():
        batch_acquisition = acquisition_function(clamped_candidates.view(-1, ndims))

    return clamped_candidates, batch_acquisition


def gen_mll_candidates_scipy(
    initial_conditions: Tensor,
    mll: Callable,
    lower_bounds: Tensor,
    upper_bounds: Tensor,
) -> Tuple[Tensor, Tensor]:
    ndims = initial_conditions.shape[-1]
    clamped_candidates = columnwise_clamp(
        X=initial_conditions, lower=lower_bounds, upper=upper_bounds
    )  # shape(num_restarts, q, ndims)
    clamped_candidates = clamped_candidates.reshape(
        (-1, ndims)
    )  # shape(num_restarts * q, ndims)
    candidates = torch.zeros_like(clamped_candidates).to(clamped_candidates)
    scipy_bounds = Bounds(
        lb=lower_bounds.detach().numpy(),
        ub=upper_bounds.detach().numpy(),
        keep_feasible=True,
    )

    def mll_np_wrapper(x: np.ndarray, fun: Callable):
        """Given a torch callable, compute value + grad given a numpy array."""
        X = torch.from_numpy(x).requires_grad_(True)  # (ndims)
        loss = fun(X)
        # compute gradient w.r.t. the inputs (does not accumulate in leaves)
        gradf = torch.autograd.grad(loss, X)[0].detach().numpy()
        if np.isnan(gradf).any():
            msg = (
                f"{np.isnan(gradf).sum()} elements of the {x.size} element "
                "gradient array `gradf` are NaN. This often indicates numerical issues."
            )
            raise RuntimeError(msg)
        fval = loss.item()
        return fval, gradf

    def neg_mll(t):
        return -mll(t)

    batch_mll_values = torch.zeros(clamped_candidates.shape[0]).to(clamped_candidates)
    for i, x0 in enumerate(clamped_candidates):
        res = minimize(
            fun=mll_np_wrapper,
            args=(neg_mll,),
            x0=x0,
            method="L-BFGS-B",
            jac=True,
            bounds=scipy_bounds,
        )
        candidates[i] = torch.from_numpy(res.x)
        batch_mll_values[i] = torch.tensor(-res.fun)
    clamped_candidates = columnwise_clamp(
        X=candidates, lower=lower_bounds, upper=upper_bounds, raise_on_violation=True
    )
    clamped_candidates = clamped_candidates.reshape(
        initial_conditions.shape
    )  # shape(num_restarts, q, ndims)

    return clamped_candidates, batch_mll_values


def _gen_mll_candidates_scipy(
    initial_conditions: Tensor,
    mll: Callable,
    lower_bounds: Tensor,
    upper_bounds: Tensor,
) -> Tuple[Tensor, Tensor]:
    ndims = initial_conditions.shape[-1]
    clamped_candidates = columnwise_clamp(
        X=initial_conditions, lower=lower_bounds, upper=upper_bounds
    )  # shape(num_restarts, q, ndims)

    shapeX = clamped_candidates.shape
    x0 = clamped_candidates.view(-1)
    bounds = make_scipy_bounds(
        X=initial_conditions, lower_bounds=lower_bounds, upper_bounds=upper_bounds
    )

    def mll_np_wrapper(x: np.ndarray, fun: Callable):
        """Given a torch callable, compute value + grad given a numpy array."""
        if np.isnan(x).any():
            raise RuntimeError(
                f"{np.isnan(x).sum()} elements of the {x.size} element array "
                f"`x` are NaN."
            )
        X = (
            torch.from_numpy(x)
            .to(initial_conditions)
            .view(-1, ndims)  # shape(num_restarts x q, ndims)
            .contiguous()
            .requires_grad_(True)
        )
        loss = torch.tensor(0.0).to(X)
        for xx in X:
            loss += fun(xx)
        # compute gradient w.r.t. the inputs (does not accumulate in leaves)
        gradf = _arrayify(torch.autograd.grad(loss, X)[0].contiguous().view(-1))
        if np.isnan(gradf).any():
            msg = (
                f"{np.isnan(gradf).sum()} elements of the {x.size} element "
                "gradient array `gradf` are NaN. This often indicates numerical issues."
            )
            if initial_conditions.dtype != torch.double:
                msg += " Consider using `dtype=torch.double`."
            raise RuntimeError(msg)
        fval = loss.item()
        return fval, gradf

    def neg_mll(t):
        return -mll(t)

    res = minimize(
        fun=mll_np_wrapper,
        args=(neg_mll,),
        x0=x0,
        method="L-BFGS-B",
        jac=True,
        bounds=bounds,
    )
    candidates = torch.from_numpy(res.x).to(initial_conditions).reshape(shapeX)

    clamped_candidates = columnwise_clamp(
        X=candidates, lower=lower_bounds, upper=upper_bounds, raise_on_violation=True
    )

    batch_mll_values = torch.zeros(shapeX[0]).to(clamped_candidates)
    for i, xx in enumerate(clamped_candidates.view(-1, ndims)):
        batch_mll_values[i] = mll(xx)

    return clamped_candidates, batch_mll_values



def test_func_time():
    import time
    import numpy as np
    from importlib import import_module, reload

    print("Start testing HDBOBenchmark.")
    import HDBOBenchmark

    print("Start testing initilization time.")
    for file in dir(HDBOBenchmark.base):
        try:
            class_name = file

            start_time = time.perf_counter_ns()
            reload(HDBOBenchmark.base)
            submodule = import_module(f"HDBOBenchmark.base.{file}")
            func = getattr(submodule, class_name)(dim=1000)
            time_cost = time.perf_counter_ns() - start_time
            print(f"Inited {class_name}, time {time_cost * 1e-6} ms.")

            x = np.random.random(func.dim)
            start_time = time.perf_counter_ns()
            for _ in range(100):
                func.evaluate(x)
            time_cost = time.perf_counter_ns() - start_time
            print(f"Run 100 * {class_name}, time {time_cost * 1e-6} ms.")
        except Exception as e:
            pass

    print("Start testing HDBOBenchmark.")
    import HDBOBenchmark

    print("Start testing initilization time.")
    start_time = time.perf_counter_ns()
    reload(HDBOBenchmark)
    time_cost = time.perf_counter_ns() - start_time
    print(f"Inited HDBOBenchmark, time {time_cost * 1e-6} ms.")
    try:
        for testset_name in HDBOBenchmark.__all__:
            for testset in getattr(HDBOBenchmark, testset_name):
                for func in testset:
                    x = np.random.random(func.dim)
                    start_time = time.perf_counter_ns()
                    for _ in range(100):
                        func.evaluate(x)
                    time_cost = time.perf_counter_ns() - start_time
                    print(f"Run 100 * {func.name}, time {time_cost * 1e-6} ms.")
    except Exception as e:
        pass

    print("Start testing botorch.")
    import torch
    import botorch.test_functions

    print("Start testing initilization time.")
    start_time = time.perf_counter_ns()
    reload(botorch.test_functions)
    time_cost = time.perf_counter_ns() - start_time
    print(f"Inited botorch.test_functions, time {time_cost * 1e-6} ms.")

    for func_name in dir(botorch.test_functions):
        try:
            func = getattr(botorch.test_functions, func_name)(dim=1000)
            x = torch.rand(func.dim)
            start_time = time.perf_counter_ns()
            for _ in range(100):
                func(x)
            time_cost = time.perf_counter_ns() - start_time
            print(f"Run 100 * {func_name}, time {time_cost * 1e-6} ms.")
        except Exception as e:
            pass

    print("Start testing HEBO.")
    import hebo.benchmarks

    print("Start testing initilization time.")
    start_time = time.perf_counter_ns()
    reload(hebo.benchmarks)
    time_cost = time.perf_counter_ns() - start_time
    print(f"Inited hebo.benchmarks, time {time_cost * 1e-6} ms.")

    for group_name in dir(hebo.benchmarks):
        group = getattr(hebo.benchmarks, group_name)
        for func_name in dir(group):
            try:
                func = getattr(group, func_name)(dim=1000)
                x = func.space.sample()
                start_time = time.perf_counter_ns()
                for _ in range(100):
                    func(x)
                time_cost = time.perf_counter_ns() - start_time
                print(f"Run 100 * {func_name}, time {time_cost * 1e-6} ms.")
            except Exception as e:
                pass


if __name__ == "__main__":
    test_func_time()



# from https://github.com/google/vizier/blob/main/vizier/_src/benchmarks/experimenters/synthetic/bbob.py
#
# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter function implementations for BBOB functions."""

import hashlib
import math
from typing import Any, Callable, Sequence
import numpy as np
from scipy import stats


## Utility Functions for BBOB.
def LambdaAlpha(alpha: float, dim: int) -> np.ndarray:
    """The BBOB LambdaAlpha matrix creation function.

    Args:
      alpha: Function parameter.
      dim: Dimension of matrix created.

    Returns:
      Diagonal matrix of dimension dim with values determined by alpha.
    """
    lambda_alpha = np.zeros([dim, dim])
    for i in range(dim):
        exp = (0.5 * (float(i) / (dim - 1))) if dim > 1 else 0.5
        lambda_alpha[i, i] = alpha**exp
    return lambda_alpha


def ArrayMap(vector: np.ndarray, fn: Callable[[float], float]) -> np.ndarray:
    """Create a new array by mapping fn() to each element of the original array.

    Args:
      vector: ndarray to be mapped.
      fn: scalar function for mapping.

    Returns:
      New ndarray be values mapped by fn.
    """
    results = np.zeros(vector.shape)
    for i, v in enumerate(vector.flat):
        results.flat[i] = fn(v)
    return results


def Tosz(element: float) -> float:
    """The BBOB T_osz function.

    Args:
      element: float input.

    Returns:
      Tosz(input).
    """
    x_carat = 0.0 if element == 0 else math.log(abs(element))
    c1 = 10.0 if element > 0 else 5.5
    c2 = 7.9 if element > 0 else 3.1
    return np.sign(element) * math.exp(
        x_carat + 0.049 * (math.sin(c1 * x_carat) + math.sin(c2 * x_carat))
    )


def Tasy(vector: np.ndarray, beta: float) -> np.ndarray:
    """The BBOB Tasy function.

    Args:
      vector: ndarray
      beta: Function parameter

    Returns:
      ndarray with values determined by beta.
    """
    dim = len(vector)
    result = np.zeros([dim, 1])
    for i, val in enumerate(vector.flat):
        if val > 0:
            t = i / (dim - 1.0) if dim > 1 else 1
            exp = 1 + beta * t * (val**0.5)
        else:
            exp = 1
        result[i] = val**exp
    return result


def SIndex(dim: int, to_sz) -> float:
    """Calculate the BBOB s_i.

    Assumes i is 0-index based.

    Args:
      dim: dimension
      to_sz: values

    Returns:
      float representing SIndex(i, d, to_sz).
    """
    s = np.zeros(
        [
            dim,
        ]
    )
    for i in range(dim):
        if dim > 1:
            s[i] = 10 ** (0.5 * (i / (dim - 1.0)))
        else:
            s[i] = 10**0.5
        if i % 2 == 0 and to_sz[i] > 0:
            s[i] *= 10
    return s


def Fpen(vector: np.ndarray) -> float:
    """The BBOB Fpen function.

    Args:
      vector: ndarray.

    Returns:
      float representing Fpen(vector).
    """
    return sum([max(0.0, (abs(x) - 5.0)) ** 2 for x in vector.flat])


def _IntSeeds(any_seeds: Sequence[Any], *, byte_length: int = 4) -> list[int]:
    """Array of integers that can be used as random state seed."""
    int_seeds = []
    for s in any_seeds:
        # Encode into 4 byte_length worth of a hexadecimal string.
        hashed = hashlib.shake_128(str(s).encode("utf-8")).hexdigest(byte_length)
        int_seeds.append(int(hashed, 16))
    return int_seeds


def _ToFloat(a: int, b: np.ndarray) -> np.ndarray:
    """Convert a%b where b is an int into a float on [-0.5, 0.5]."""
    return (np.int64(a) % b) / np.float64(b) - 0.5


def _R(dim: int, seed: int, *moreseeds: Any) -> np.ndarray:
    """Returns an orthonormal rotation matrix.

    Args:
      dim: size of the resulting matrix.
      seed: int seed. If set to 0, this function returns an identity matrix
        regardless of *moreseeds.
      *moreseeds: Additional parameters to include in the hash. Arguments are
        converted to strings first.

    Returns:
      Array of shape (dim, dim), representing a rotation matrix.
    """
    if seed == 0:
        return np.identity(dim)
    rng = np.random.default_rng(_IntSeeds(((seed, dim) + moreseeds)))
    return stats.special_ortho_group.rvs(dim, random_state=rng)


## BBOB Functions.
def Sphere(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Sphere function."""
    del seed
    return float(np.sum(arr * arr))


def Rastrigin(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Rastrigin function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z = np.matmul(_R(dim, seed, b"R"), arr)
    z = Tasy(ArrayMap(z, Tosz), 0.2)
    z = np.matmul(_R(dim, seed, b"Q"), z)
    z = np.matmul(LambdaAlpha(10.0, dim), z)
    z = np.matmul(_R(dim, seed, b"R"), z)
    return float(10 * (dim - np.sum(np.cos(2 * math.pi * z))) + np.sum(z * z, axis=0))


def BuecheRastrigin(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB BuecheRastrigin function."""
    del seed
    dim = len(arr)
    arr.shape = (dim, 1)
    t = ArrayMap(arr, Tosz)
    l = SIndex(dim, arr) * t.flat

    term1 = 10 * (dim - np.sum(np.cos(2 * math.pi * l), axis=0))
    term2 = np.sum(l * l, axis=0)
    term3 = 100 * Fpen(arr)
    return float(term1 + term2 + term3)


def LinearSlope(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB LinearSlope function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    r = _R(dim, seed, b"R")
    z = np.matmul(r, arr)
    result = 0.0
    for i in range(dim):
        s = 10 ** (i / float(dim - 1) if dim > 1 else 1)
        z_opt = 5 * np.sum(np.abs(r[i, :]))
        result += float(s * (z_opt - z[i]))
    return result


def AttractiveSector(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Attractive Sector function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    x_opt = np.array([1 if i % 2 == 0 else -1 for i in range(dim)])
    x_opt.shape = (dim, 1)
    z_vec = np.matmul(_R(dim, seed, b"R"), arr - x_opt)
    z_vec = np.matmul(LambdaAlpha(10.0, dim), z_vec)
    z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

    result = 0.0
    for i in range(dim):
        z = z_vec[i, 0]
        s = 100 if z * x_opt[i] > 0 else 1
        result += (s * z) ** 2

    return math.pow(Tosz(result), 0.9)


def StepEllipsoidal(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB StepEllipsoidal function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z_hat = np.matmul(_R(dim, seed, b"R"), arr)
    z_hat = np.matmul(LambdaAlpha(10.0, dim), z_hat)
    z_tilde = np.array(
        [
            math.floor(0.5 + z) if (z > 0.5) else (math.floor(0.5 + 10 * z) / 10)
            for z in z_hat.flat
        ]
    )
    z_tilde = np.matmul(_R(dim, seed, b"Q"), z_tilde)
    s = 0.0
    for i, val in enumerate(z_tilde):
        exponent = 2.0 * float(i) / (dim - 1.0) if dim > 1.0 else 2.0
        s += 10.0**exponent * val**2
    value = max(abs(z_hat[0, 0]) / 1000, s)
    return 0.1 * value + Fpen(arr)


def RosenbrockRotated(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB RosenbrockRotated function."""
    dim = len(arr)
    r_x = np.matmul(_R(dim, seed, b"R"), arr)
    z = max(1.0, (dim**0.5) / 8.0) * r_x + 0.5 * np.ones((dim,))
    return float(
        sum(
            [
                100.0 * (z[i] ** 2 - z[i + 1]) ** 2 + (z[i] - 1) ** 2
                for i in range(dim - 1)
            ]
        )
    )


def Ellipsoidal(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Ellipsoidal function."""
    del seed
    dim = len(arr)
    arr.shape = (dim, 1)
    z_vec = ArrayMap(arr, Tosz)
    s = 0.0
    for i in range(dim):
        exp = 6.0 * i / (dim - 1) if dim > 1 else 6.0
        s += float(10**exp * z_vec[i] * z_vec[i])
    return s


def Discus(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Discus function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    r_x = np.matmul(_R(dim, seed, b"R"), arr)
    z_vec = ArrayMap(r_x, Tosz)
    return float(10**6 * z_vec[0] * z_vec[0]) + sum([z * z for z in z_vec[1:].flat])


def BentCigar(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB BentCigar function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z_vec = np.matmul(_R(dim, seed, b"R"), arr)
    z_vec = Tasy(z_vec, 0.5)
    z_vec = np.matmul(_R(dim, seed, b"R"), z_vec)
    return float(z_vec[0] ** 2) + 10**6 * np.sum(z_vec[1:] ** 2)


def SharpRidge(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB SharpRidge function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z_vec = np.matmul(_R(dim, seed, b"R"), arr)
    z_vec = np.matmul(LambdaAlpha(10, dim), z_vec)
    z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)
    return z_vec[0, 0] ** 2 + 100 * np.sum(z_vec[1:] ** 2) ** 0.5


def DifferentPowers(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB DifferentPowers function."""
    dim = len(arr)
    z = np.matmul(_R(dim, seed, b"R"), arr)
    s = 0.0
    for i in range(dim):
        exp = 2 + 4 * i / (dim - 1) if dim > 1 else 6
        s += abs(z[i]) ** exp
    return s**0.5


def Weierstrass(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Weierstrass function."""
    k_order = 12
    dim = len(arr)
    arr.shape = (dim, 1)
    z = np.matmul(_R(dim, seed, b"R"), arr)
    z = ArrayMap(z, Tosz)
    z = np.matmul(_R(dim, seed, b"Q"), z)
    z = np.matmul(LambdaAlpha(1.0 / 100.0, dim), z)
    f0 = sum([0.5**k * math.cos(math.pi * 3**k) for k in range(k_order)])

    s = 0.0
    for i in range(dim):
        for k in range(k_order):
            s += 0.5**k * math.cos(2 * math.pi * (3**k) * (z[i] + 0.5))

    return float(10 * (s / dim - f0) ** 3) + 10 * Fpen(arr) / dim


def SchaffersF7(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Weierstrass function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    if dim == 1:
        return 0.0
    z = np.matmul(_R(dim, seed, b"R"), arr)
    z = Tasy(z, 0.5)
    z = np.matmul(_R(dim, seed, b"Q"), z)
    z = np.matmul(LambdaAlpha(10.0, dim), z)

    s_arr = np.zeros(dim - 1)
    for i in range(dim - 1):
        s_arr[i] = float((z[i] ** 2 + z[i + 1] ** 2) ** 0.5)
    s = 0.0
    for i in range(dim - 1):
        s += s_arr[i] ** 0.5 + (s_arr[i] ** 0.5) * math.sin(50 * s_arr[i] ** 0.2) ** 2

    return (s / (dim - 1.0)) ** 2 + 10 * Fpen(arr)


def SchaffersF7IllConditioned(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB SchaffersF7 Ill Conditioned."""
    dim = len(arr)
    arr.shape = (dim, 1)
    if dim == 1:
        return 0.0
    z = np.matmul(_R(dim, seed, b"R"), arr)
    z = Tasy(z, 0.5)
    z = np.matmul(_R(dim, seed, b"Q"), z)
    z = np.matmul(LambdaAlpha(1000.0, dim), z)

    s_arr = np.zeros(dim - 1)
    for i in range(dim - 1):
        s_arr[i] = float((z[i] ** 2 + z[i + 1] ** 2) ** 0.5)
    s = 0.0
    for i in range(dim - 1):
        s += s_arr[i] ** 0.5 + (s_arr[i] ** 0.5) * math.sin(50 * s_arr[i] ** 0.2) ** 2

    return (s / (dim - 1.0)) ** 2 + 10 * Fpen(arr)


def GriewankRosenbrock(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB GriewankRosenbrock function."""
    dim = len(arr)
    r_x = np.matmul(_R(dim, seed, b"R"), arr)
    # Slightly off BBOB documentation in order to center optima at origin.
    # Should be: max(1.0, (dim**0.5) / 8.0) * r_x + 0.5 * np.ones((dim,)).
    z_arr = max(1.0, (dim**0.5) / 8.0) * r_x + np.ones((dim,))
    s_arr = np.zeros(dim)
    for i in range(dim - 1):
        s_arr[i] = 100.0 * (z_arr[i] ** 2 - z_arr[i + 1]) ** 2 + (z_arr[i] - 1) ** 2

    total = 0.0
    for i in range(dim - 1):
        total += s_arr[i] / 4000.0 - math.cos(s_arr[i])

    return (10.0 * total) / (dim - 1) + 10


def Schwefel(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Schwefel function."""
    del seed
    dim = len(arr)
    bernoulli_arr = np.array([pow(-1, i + 1) for i in range(dim)])
    x_opt = 4.2096874633 / 2.0 * bernoulli_arr
    x_hat = 2.0 * (bernoulli_arr * arr)  # Element-wise multiplication

    z_hat = np.zeros([dim, 1])
    z_hat[0, 0] = x_hat[0]
    for i in range(1, dim):
        z_hat[i, 0] = x_hat[i] + 0.25 * (x_hat[i - 1] - 2 * abs(x_opt[i - 1]))

    x_opt.shape = (dim, 1)
    z_vec = 100 * (
        np.matmul(LambdaAlpha(10, dim), z_hat - 2 * abs(x_opt)) + 2 * abs(x_opt)
    )

    total = sum([z * math.sin(abs(z) ** 0.5) for z in z_vec.flat])

    return -(total / (100.0 * dim)) + 4.189828872724339 + 100 * Fpen(z_vec / 100)


def Katsuura(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Katsuura function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    r_x = np.matmul(_R(dim, seed, b"R"), arr)
    z_vec = np.matmul(LambdaAlpha(100.0, dim), r_x)
    z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

    prod = 1.0
    for i in range(dim):
        s = 0.0
        for j in range(1, 33):
            s += abs(2**j * z_vec[i, 0] - round(2**j * z_vec[i, 0])) / 2**j
        prod *= (1 + (i + 1) * s) ** (10.0 / dim**1.2)

    return (10.0 / dim**2) * prod - 10.0 / dim**2 + Fpen(arr)


def Lunacek(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Lunacek function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    mu0 = 2.5
    s = 1.0 - 1.0 / (2.0 * (dim + 20.0) ** 0.5 - 8.2)
    mu1 = -(((mu0**2 - 1) / s) ** 0.5)

    x_opt = np.array([mu0 / 2] * dim)
    x_hat = np.array([2 * arr[i, 0] * np.sign(x_opt[i]) for i in range(dim)])
    x_vec = x_hat - mu0
    x_vec.shape = (dim, 1)
    x_vec = np.matmul(_R(dim, seed, b"R"), x_vec)
    z_vec = np.matmul(LambdaAlpha(100, dim), x_vec)
    z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

    s1 = sum([(val - mu0) ** 2 for val in x_hat])
    s2 = sum([(val - mu1) ** 2 for val in x_hat])
    s3 = sum([math.cos(2 * math.pi * z) for z in z_vec.flat])
    return min(s1, dim + s * s2) + 10.0 * (dim - s3) + 10**4 * Fpen(arr)


def Gallagher101Me(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Gallagher101 function."""
    dim = len(arr)
    arr.shape = (dim, 1)

    num_optima = 101
    optima_list = [np.zeros([dim, 1])]
    for i in range(num_optima - 1):
        vec = np.zeros([dim, 1])
        for j in range(dim):
            alpha = (i * dim + j + 1.0) / (dim * num_optima + 2.0)
            assert alpha > 0
            assert alpha < 1
            vec[j, 0] = -5 + 10 * alpha
        optima_list.append(vec)

    c_list = [LambdaAlpha(1000, dim)]
    for i in range(num_optima - 1):
        alpha = 1000.0 ** (2.0 * (i) / (num_optima - 2))
        c_mat = LambdaAlpha(alpha, dim) / (alpha**0.25)
        c_list.append(c_mat)

    rotation = _R(dim, seed, b"R")
    max_value = -1.0
    for i in range(num_optima):
        w = 10 if i == 0 else (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0))
        diff = np.matmul(rotation, arr - optima_list[i])
        e = np.matmul(diff.transpose(), np.matmul(c_list[i], diff))
        max_value = max(max_value, w * math.exp(-float(e) / (2.0 * dim)))

    return Tosz(10.0 - max_value) ** 2 + Fpen(arr)


def Gallagher21Me(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Gallagher21 function."""
    dim = len(arr)
    arr.shape = (dim, 1)

    num_optima = 21
    optima_list = [np.zeros([dim, 1])]
    for i in range(num_optima - 1):
        vec = np.zeros([dim, 1])
        for j in range(dim):
            alpha = (i * dim + j + 1.0) / (dim * num_optima + 2.0)
            assert alpha > 0
            assert alpha < 1
            vec[j, 0] = -5 + 10 * alpha
        optima_list.append(vec)

    c_list = [LambdaAlpha(1000, dim)]
    for i in range(num_optima - 1):
        alpha = 1000.0 ** (2.0 * (i) / (num_optima - 2))
        c_mat = LambdaAlpha(alpha, dim) / (alpha**0.25)
        c_list.append(c_mat)

    rotation = _R(dim, seed, b"R")
    max_value = -1.0
    for i in range(num_optima):
        w = 10 if i == 0 else (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0))
        diff = np.matmul(rotation, arr - optima_list[i])
        e = np.matmul(diff.transpose(), np.matmul(c_list[i], diff))
        max_value = max(max_value, w * math.exp(-float(e) / (2.0 * dim)))

    return Tosz(10.0 - max_value) ** 2 + Fpen(arr)


## Additional BBOB-like functions to test exploration.


def NegativeSphere(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for BBOB Sphere function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z = np.matmul(_R(dim, seed, b"R"), arr)
    return float(100 + np.sum(z * z) - 2 * (z[0] ** 2))


def NegativeMinDifference(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for NegativeMinDifference function."""
    dim = len(arr)
    arr.shape = (dim, 1)
    z = np.matmul(_R(dim, seed, b"R"), arr)
    min_difference = 10000
    for i in range(len(z) - 1):
        min_difference = min(min_difference, z[i + 1] - z[i])
    return 10.0 - float(min_difference) + 1e-8 * float(sum(arr))


def FonsecaFleming(arr: np.ndarray, seed: int = 0) -> float:
    """Implementation for FonsecaFleming function."""
    del seed
    return 1.0 - float(np.exp(-np.sum(arr * arr)))



from . import bbob
from .wrapper import VizierBBOBWrapper

bbob_func_names = (
    "Rastrigin",
    "LinearSlope",
    "AttractiveSector",
    "StepEllipsoidal",
    "RosenbrockRotated",
    "Discus",
    "BentCigar",
    "SharpRidge",
    "DifferentPowers",
    "Weierstrass",
    "SchaffersF7",
    "SchaffersF7IllConditioned",
    "GriewankRosenbrock",
    "Katsuura",
    "Lunacek",
    "Gallagher101Me",
    "Gallagher21Me",
    "NegativeSphere",
    "NegativeMinDifference",
)

bbob_dim = 10
bbob_func_dict = dict()  # minimize
for name in bbob_func_names:
    bbob_func_dict[name] = VizierBBOBWrapper(func=getattr(bbob, name), dim=bbob_dim)



import torch


class _HyperParam:
    def __init__(self, ndims: int, num_groups: int, add_remaining_dims=True):
        # First set the Decomposition
        self.decomp_strategy = "partialLearn"
        ndims_per_group = ndims // num_groups
        rem_dims = ndims - ndims_per_group * num_groups

        decomp = ndims_per_group * torch.ones(num_groups).long()

        if rem_dims != 0:
            decomp = torch.hstack(
                (
                    decomp,
                    torch.tensor([rem_dims]),
                )
            )
            num_groups += 1

        self.decomp = decomp
        self.num_groups = num_groups
        # Hyper Parameters for optimization
        # Utility (Acquisition) Function - use UCB as default
        self.utility_func = "UCB"
        # The not-so-important hyper parameters
        self.opt_pt_std_threshold = 0.002

        # GP Hyper parameter for Regression
        # The Common Mean Function
        self.common_mean_func = lambda arg: torch.zeros(arg.shape[0]).to(
            arg
        )  # By default will use all zeros.
        # The Mean function for the individual GPs
        self.mean_funcs = lambda arg: torch.zeros(arg.shape[0]).to(arg)
        # Common noise parameters
        self.common_noise = 1e-2
        # Individual noise
        self.noises = torch.zeros(num_groups, dtype=torch.double)
        # Scale parameters
        self.fix_pr = False
        self.use_same_pr = True
        self.sigma_pr_range = torch.DoubleTensor([0.03, 30.0])
        self.sigma_pr_ranges = []  # use same prior by default
        # Bandwidth parameters
        self.use_fixed_bandwidth = False
        self.fix_sm = False
        self.use_same_sm = True
        self.al_bw_lb = 1e-5
        self.al_bw_ub = 5


class HyperParam:
    def __init__(self, ndims, ndims_per_group, add_remaining_dims=True):
        # First set the Decomposition
        self.decomp_strategy = "partialLearn"
        if add_remaining_dims:
            num_groups = torch.ceil(torch.tensor(ndims / ndims_per_group)).long()
            rem_dims = ndims - ndims_per_group * (num_groups - 1)
            if rem_dims == 0:
                add_remaining_dims = False
        else:
            num_groups = ndims // ndims_per_group

        # Determine the decomposition accordingly.
        if self.decomp_strategy == "known":
            decomp = []
            for i in range(num_groups):
                decomp.append(
                    torch.arange(
                        i * ndims_per_group,
                        min((i + 1) * ndims_per_group, ndims),
                        dtype=torch.long,
                    )
                )
        elif add_remaining_dims:
            decomp = torch.hstack(
                (ndims_per_group * torch.ones(num_groups - 1).long(), rem_dims)
            )
        else:
            raise NotImplementedError
        self.decomp = decomp
        self.num_groups = num_groups
        # Hyper Parameters for optimization
        # Utility (Acquisition) Function - use UCB as default
        self.utility_func = "UCB"
        # The not-so-important hyper parameters
        self.opt_pt_std_threshold = 0.002

        # GP Hyper parameter for Regression
        # The Common Mean Function
        self.common_mean_func = lambda arg: torch.zeros(arg.shape[0]).to(
            arg
        )  # By default will use all zeros.
        # The Mean function for the individual GPs
        self.mean_funcs = lambda arg: torch.zeros(arg.shape[0]).to(arg)
        # Common noise parameters
        self.common_noise = 1e-2
        # Individual noise
        self.noises = torch.zeros(num_groups, dtype=torch.double)
        # Scale parameters
        self.fix_pr = False
        self.use_same_pr = True
        self.sigma_pr_range = torch.DoubleTensor([0.03, 30.0])
        self.sigma_pr_ranges = []  # use same prior by default
        # Bandwidth parameters
        self.use_fixed_bandwidth = False
        self.fix_sm = False
        self.use_same_sm = True
        self.al_bw_lb = 1e-5
        self.al_bw_ub = 5



from HDBOBenchmark.utils.logging import logger
from math import log2
from numpy import ndarray
from tqdm import tqdm
import torch
from typing import Callable
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.addGPRegression import (
    add_gp_regression,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.addGPDecompMargLikelihood import (
    add_gp_decomp_marg_likelihood,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.utils.projectToRectangle import (
    project_to_rectangle,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_utils.sampling_torch import (
    draw_sobol_samples,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.BOLibkky.getNextQueryPoint import (
    get_next_query_pt,
)


def add_gp_bo(
    oracle: Callable,
    bounds: ndarray,
    num_iterations: int,
    params,
):
    """Implements Gaussian Process Bandits/ Bayesian Optimization using Additive Gaussian Processes.
    See ICML 2015 parper: "High Dimensional Bayesian Optimization and Bandits via Additive Models". K.Kandasamy,
    J.Schneider, B.Poczos
    Args:
        oracle: A function handle for the function you wish to optimize.
        bounds: A 'numDims x 2' ndarray specifying the lower and upper bound of each dimension.
        num_iterations: The number of GPB/BO iterations.
        params: A sturcture specifying the various hyper parameters for optimization. If you wish
            to use default settings, pass an empty struct. Otherwise, see demoCustomise.py to see how to set each hyper
            parameter. Also see below.
    Returns:
        maxVal: The maximum queried value of the function.
        maxPt: The queried point with the maximum value.
        boQueries: A matrix indicating the points at which the algorithm queried.
        boVals: A vector of the query values.
        history: A vector of the maximum value obtained up until that iteration.

    params should have a field called decompStrategy: It should be one of 'known', 'learn', 'random' and 'partialLearn'.
    'known': The decomposition is known and given in decomp. We will optimize according to this.
    'learn': The decomposition is unknown and should be learned.
    'random': Randomly pick a partition at each iteration.
    'partialLearn': Partially learn the decomposition at each iteration by trying out a few and picking the best.
    The default is partialLearn and is the best option if you don't know much about your function.
    """

    # Prelims
    if isinstance(bounds, ndarray):
        bounds = torch.from_numpy(bounds)
    ndims = bounds.shape[0]
    max_threshold_exceeds, num_iter_param_relearn = 5, 25
    decomp = params.decomp
    # The Decomposition
    if params.decomp_strategy == "known":
        decomposition = decomp
        num_groups = len(decomposition)
        # do some diagnostics on the decomposition and print them out
        # relevant_coords = torch.hstack(decomposition)
        # num_relevant_coords = len(relevant_coords)
        # if num_relevant_coords != len(torch.unique(relevant_coords)):
        #     raise Exception('The same coordinate cannot appear in different groups')
        # logger.info('# Groups: %d, %d/%d coordinates are relevant\n' % (num_groups, num_relevant_coords, ndims))
    elif hasattr(decomp, "m"):
        # Now decomposition should have two fields d and m
        num_groups = decomp.m
    else:  # in this case the decomposition is given.
        num_groups = len(decomp)

    # Initialization points
    num_init_pts = 2 ** int(log2(num_iterations // 10))
    # num_init_pts = 10
    logger.info("Obtaining %d Points for Initialization.\n" % num_init_pts)
    init_pts = draw_sobol_samples(
        bounds=bounds.T, n=num_init_pts, q=1
    ).squeeze()  # shape(num_init, ndims)

    """init_pts = torch.repeat_interleave(torch.arange(1, num_init_pts+1, dtype=bounds.dtype).reshape(-1, 1) / num_init_pts, ndims, dim=1)
    init_pts = init_pts * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]"""

    init_vals = torch.tensor([oracle(pt) for pt in init_pts], dtype=init_pts.dtype)

    num_iterations -= num_init_pts

    # The Bandwidth
    # This BO algorithm will set the bandwidth via its own procedure
    al_bw_lb = params.al_bw_lb
    al_bw_ub = params.al_bw_ub
    # Set an initial bandwidth. This will change as the algorithm progresses
    al_curr_bw = al_bw_ub

    # Define the following before proceeding
    bo_queries = torch.vstack(
        (init_pts, torch.zeros(num_iterations, ndims).to(init_pts))
    )
    bo_vals = torch.hstack((init_vals, torch.zeros(num_iterations).to(init_vals)))
    best_x, best_y, history = (
        init_pts[0],
        init_vals[0],
        torch.zeros(num_init_pts + num_iterations).to(bo_vals),
    )
    history[0] = best_y
    for i in torch.arange(1, len(init_vals)):
        if init_vals[i] > best_y:
            best_y = init_vals[i]
            best_x = init_pts[i]
        history[i] = best_y

    thresh_exceeded_counter = 0

    logger.info("Performing BO (dim = %d)\n" % ndims)
    pbar = tqdm(total=num_iterations)
    for bo_iter in range(num_iterations):
        if (bo_iter + 1) % num_iter_param_relearn == 0:
            pbar.set_postfix({"MaxVal": "%0.4f" % best_y})

        # Prelims
        num_x = num_init_pts + bo_iter
        input_x = bo_queries[:num_x]
        output_y = bo_vals[:num_x]
        # standardize to mean 0 and var 1
        y_mean = torch.mean(output_y, dim=0)  # shape(num) => ()
        y_std = torch.std(output_y, dim=0)  # shape()
        train_y = (output_y - y_mean) / y_std  # shape(num)

        train_x = input_x
        scale_bounds = bounds
        """min-max scale
        min_x = torch.min(input_x, dim=0, keepdims=True).values
        max_x = torch.max(input_x, dim=0, keepdims=True).values  # shape(1, ndims)
        train_x = (input_x - min_x) / (max_x - min_x)
        scale_bounds = (bounds - min_x.T) / (max_x.T - min_x.T)
        """

        # First redefine ranges for the GP bandwidth if needed
        if (not params.use_fixed_bandwidth) and (
            bo_iter % num_iter_param_relearn == 0
            or thresh_exceeded_counter == max_threshold_exceeds
        ):
            if thresh_exceeded_counter == max_threshold_exceeds:
                al_bw_ub = max(al_bw_lb, 0.9 * al_curr_bw)
                thresh_exceeded_counter = 0
                logger.warning(
                    "Threshold Exceeded %d times - Reducing BW\n"
                    % max_threshold_exceeds
                )
            else:
                pass
            # Define the BW range for addGPMargLikelihood
            if al_bw_ub == al_bw_lb:
                params.fix_sm = True
                params.sigma_pr_ranges = al_bw_lb * torch.ones(num_groups)
            else:
                params.fix_sm = False
                # Use same bandwidth for now.
                params.use_same_sm = True
                params.sigma_sm_range = torch.DoubleTensor([al_bw_lb, al_bw_ub])

            # Obtain the optimal GP parameters
            if params.decomp_strategy != "stoch1":
                (
                    al_curr_bws,
                    al_curr_scales,
                    _,
                    learned_decomp,
                    marg_like_val,
                ) = add_gp_decomp_marg_likelihood(train_x, train_y, decomp, params)
                al_curr_bw = al_curr_bws[0]  # modify to allow different bandwidths
            else:
                raise NotImplementedError

        # If stochastic pick a current GP
        if params.decomp_strategy != "stoch1":
            params.sigma_sms = al_curr_bws
            params.sigma_prs = al_curr_scales
            curr_iter_decomp = learned_decomp
        else:
            raise NotImplementedError

        # Now build the GP
        add_gp_regression(train_x, train_y, curr_iter_decomp, params)
        # Now obtain the next point
        candidate, _, candidate_std = get_next_query_pt(
            train_x, train_y, params, curr_iter_decomp, scale_bounds
        )
        # If it is too close, perturb it a bit
        while (
            torch.min(
                torch.sqrt(torch.sum((train_x - candidate.unsqueeze(0)) ** 2, dim=1))
            )
            / al_curr_bw
            < 1e-10
        ):
            logger.debug("The candidate is too close to training set, perturb it a bit")
            candidate = project_to_rectangle(
                candidate + 0.1 * al_curr_bw * torch.randn(ndims), scale_bounds
            )
        # Determine the current best point
        """invert min-max scale
        candidate = candidate * (max_x - min_x) + min_x
        """
        candidate = candidate.squeeze()
        value = oracle(candidate)
        if value > best_y:
            best_y = value
            best_x = candidate

        bo_queries[num_init_pts + bo_iter] = candidate
        bo_vals[num_init_pts + bo_iter] = value

        # Check if next_point_std is too small
        if candidate_std < params.opt_pt_std_threshold:
            thresh_exceeded_counter += 1
        else:
            thresh_exceeded_counter = 0
        # Store the current best value
        history[bo_iter + num_init_pts] = best_y
        pbar.update()

    return (
        best_y,
        best_x.numpy(),
        bo_queries.numpy(),
        bo_vals.numpy(),
        history.numpy(),
    )



"""
==================================================
MIPLIB 2017
URL: https://miplib.zib.de/index.html
==================================================
"""
from collections import defaultdict
from pathlib import Path
import numpy as np
from pyscipopt import Model, __version__ as ver_pyscipopt
from HDBOBenchmark.funcs.base.FunctionBase import TestFunction
from .util import download_miplib, get_solutions
from .parse_para_utils import parse_scip_para


class MPSModel(TestFunction):
    def __init__(
        self,
        mps_path,
        solu_path="miplib2017-v26.solu",
        asset_path=r"assets/miplib/",
        dim=None,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=False,
        solver="scip",
        t_max=7200,  # 2 hours
        seed=None,
        *args,
        **kwargs
    ):
        """
        Args:
            mps_path: the relative path of .mps task.
            solu_path: the relative path of .solu.
            asset_path: the asset path.
            solver: "scip" to optimize the hyperparameters of SCIP solver, others to optimize the original objective of mps.
            t_max: max solve time(seconds) for solvers.
        """
        assert (
            mps_path is not None
        ), "You need to read .mps file to load the problem first!"
        TestFunction.__init__(self)
        self.asset_path = Path(asset_path)
        self.mps_path = mps_path
        self.solu_path = self.asset_path / solu_path
        self.check_assets()
        self.solver = solver
        self.t_max = t_max
        self.seed = seed
        self.getMPSModel(mps_path, t_max, seed)

        self._x_bound = np.array([[-1, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        self._min_pt = None
        self._if_torch = False
        if solver == "scip":
            self.init_SCIP()
        else:
            self.init_raw()
        if dim is not None:
            self._dim = min(dim, self.dim)
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def check_assets(self):
        if not (self.asset_path / "revised-submissions").exists():
            download_miplib(path=self.asset_path.resolve(), target="benchmark.zip")
        download_miplib(path=self.asset_path.resolve(), target="miplib2017-v26.solu")

    def getMPSModel(self, mps_path=None, t_max=None, seed=None):
        mps_path = self.mps_path if mps_path is None else mps_path
        t_max = self.t_max if t_max is None else t_max
        seed = self.seed if seed is None else seed
        mpsModel = Model()
        mpsModel.hideOutput(quiet=True)
        mpsModel.readProblem((self.asset_path / mps_path).resolve())
        mpsModel.setParam("display/verblevel", 0)
        mpsModel.setParam("limits/time", t_max)
        if seed:
            mpsModel.setParam("randomization/randomseedshift", seed)
        self.mpsModel = mpsModel

        self._name = self.mpsModel.getProbName()
        if self.solver == "scip":
            self._dim = 135
        else:
            self._dim = self.mpsModel.getNVars()
        # self.if_minium = True if mpsModel.getObjectiveSense() == "minimize" else False
        y_scale_dict = defaultdict(
            lambda: 100, {"markshare_4_0": 1e18, "enlight_hard": 1e31}
        )
        self.y_scale = y_scale_dict[self.name]
        self.solution = get_solutions(self.solu_path.resolve())[self.name]
        self._min_val = float(self.solution["answer"])

    def init_raw(self):
        self.__vars = []
        self.objective_coefficient_list = []
        self.constraint_param_list = []
        # parse variables
        variables = self.mpsModel.getVars()
        type_map = {
            "BINARY": "bool",
            "INTEGER": "int",
            "IMPLINT": "int",
            "CONTINUOUS": "continuous",
        }
        var_list = []
        objective_coefficient_list = []
        for var in variables:
            vtype = var.vtype()
            lb = np.NINF if var.getLbOriginal() == -1e20 else var.getLbOriginal()
            ub = np.Inf if var.getUbOriginal() == 1e20 else var.getUbOriginal()
            index = var.getIndex()
            objective_coefficient = var.getObj()

            var_list.append(
                {
                    "name": str(var),
                    "type": type_map[vtype],
                    "lb": lb,
                    "ub": ub,
                    "index": index,
                }
            )
            self._x_bound[index] = np.array([lb, ub])

            objective_coefficient_list.append(objective_coefficient)

        self.__vars = sorted(var_list, key=lambda d: d["index"])
        self.objective_coefficient_list = objective_coefficient_list
        self.dict_varname_num = dict.fromkeys(
            [var["name"] for var in self.variables], 0.0
        )

        # parse constraints
        conss = self.mpsModel.getConss()
        constraint_param_list = (
            []
        )  # each item is a dict, {'x1': xxx, 'x2': xxx, ... , 'lhs': xxx, 'rhs': xxx}
        unconstraint_conss = []
        for cons in conss:
            item = self.mpsModel.getValsLinear(cons) if cons.isLinear() else {}
            item["lhs"] = self.mpsModel.getLhs(cons)  # lhs < xxx < rhs
            item["rhs"] = self.mpsModel.getRhs(cons)
            constraint_param_list.append(item)
            coefficient_dict = {
                var.name: var.getLPSol() for var in self.mpsModel.getVars()
            }
            coefficient_dict.update(
                self.mpsModel.getValsLinear(cons) if cons.isLinear() else {}
            )
            unconstraint_conss.append(
                {"coefficient": list(coefficient_dict.values()), "rhs": item["rhs"]}
            )
            unconstraint_conss.append(
                {
                    "coefficient": (
                        -1 * np.array(list(coefficient_dict.values()))
                    ).tolist(),
                    "rhs": -1 * item["lhs"],
                }
            )
        self.constraint_param_list = constraint_param_list
        self.unconstraint_conss = unconstraint_conss

        # bound
        self._x_bound = np.array([[-1, 1]] * self.dim)  # shape(dim, 2)
        self._x_bound_is_pow = np.array([[0, 0]] * self.dim)  # shape(dim, 2)
        for msg_var in self.variables:
            lb, ub = msg_var["lb"], msg_var["ub"]
            if lb == -np.inf:
                self._x_bound[msg_var["index"], 0] = -100 + np.min(ub, 0)
                self._x_bound_is_pow[msg_var["index"], 0] = 1
            else:
                self._x_bound[msg_var["index"], 0] = lb
            if ub == np.inf:
                self._x_bound[msg_var["index"], 1] = 100 + np.max(lb, 0)
                self._x_bound_is_pow[msg_var["index"], 1] = 1
            else:
                self._x_bound[msg_var["index"], 1] = ub
        self._evaluate = self._evaluate_raw

    def init_SCIP(self):
        design_space, _ = parse_scip_para(Path(__file__).parent / "paras_extended.csv")
        self.params = list(design_space.paras.values())
        self._dim = len(self.params)
        self._x_bound = np.array([[-1, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        self.lb, self.ub = np.empty(self.dim), np.empty(self.dim)
        for i, param in enumerate(self.params):
            self._x_bound[i, 0] = param.lb
            self._x_bound[i, 1] = param.ub
            if param.is_discrete:
                self._x_bound[i, 1] += 1
        self._evaluate = self._evaluate_by_SCIPsolver
        # For time result
        self._min_val = 0
        self.y_scale = float(self.t_max * 2)

    def _evaluate_unconstraint_conss(self, x):
        M = 80
        loss = 0
        for unconstraint_cons in self.unconstraint_conss:
            fn = np.maximum if self.if_minium else np.minimum
            loss += fn(
                np.sum(unconstraint_cons["coefficient"] * x) - unconstraint_cons["rhs"],
                0,
            )
        return M * loss

    def _evaluate_raw(self, x):
        if len(x.shape) > 1:
            x = x[0]
        for i in range(self.dim):
            if self._x_bound_is_pow[i, 0] == 1:
                bias = np.min(self._x_bound[i, 1], 0)
                if x[i] < bias:
                    x[i] = -np.float_power(2, bias - x[i]) + bias
            if self._x_bound_is_pow[i, 1] == 1:
                bias = np.max(self._x_bound[i, 0], 0)
                if x[i] > bias:
                    x[i] = np.float_power(2, x[i] - bias) + bias
            if self.variables[i]["type"] == "bool":
                x[i] = 0 if x[i] < 0.5 else 1

        sum_original_objective = np.sum(
            self.objective_coefficient * x
        )  # Element-wise product
        sum_constraint = self._evaluate_unconstraint_conss(x)
        return sum_original_objective + sum_constraint

    def _evaluate_by_SCIPsolver(self, x_batch):
        if not isinstance(x_batch, np.ndarray):
            x_batch = np.array(x_batch)
        if x_batch.ndim == 1:
            x_batch = x_batch[np.newaxis]
        result = np.zeros(x_batch.shape[0])
        for i_batch, x in enumerate(x_batch):
            # setParam
            self.getMPSModel()
            mpsModel = self.mpsModel
            assert len(x) >= self.dim
            for i, param in enumerate(self.params):
                if i >= self.dim:
                    break
                k = param.name
                v = x[i]
                if not isinstance(v, str):
                    if param.is_discrete:
                        v = int(v)
                    if param.is_categorical:
                        v = param.categories[min(v, param.ub)]
                try:
                    mpsModel.setParam(k, v)
                except KeyError:
                    k = k.split("/")[-1]
                    if ver_pyscipopt <= "3.5.0":
                        k = "separating/" + k
                    else:
                        k = "cutselection/hybrid/" + k
                    mpsModel.setParam(k, v)

            mpsModel.optimize()
            status = mpsModel.getStatus()
            if status == "optimal":
                opt_time = mpsModel.getSolvingTime()
                obj_val = mpsModel.getObjVal()
            else:
                if status == "infeasible":
                    opt_time = self.y_scale
                    obj_val = 1e30
                else:
                    try:
                        obj_val = mpsModel.getObjVal()
                        if abs(obj_val - float(self.solution["answer"])) < 1e-12:
                            opt_time = min(mpsModel.getSolvingTime(), self.t_max)
                        else:
                            opt_time = self.y_scale
                    except:
                        obj_val = 1e30
                        opt_time = self.y_scale
            result[i_batch] = opt_time
        return result
        """
        opt_time = mpsModel.getSolvingTime()
        n_sols = mpsModel.getNSols()
        sol = mpsModel.getBestSol()
        primal_bound = mpsModel.getPrimalbound()
        dual_bound = mpsModel.getDualbound()
        gap = mpsModel.getGap()
        """

    @property
    def variables(self):
        return self.__vars

    @property
    def objective_coefficient(self):
        return self.objective_coefficient_list

    @property
    def constraint_params(self):
        # Each item is a dict: {'x1': xxx, 'x2': xxx, ... , 'lhs': xxx, 'rhs': xxx}
        return self.constraint_param_list



import torch
from torch import Tensor
from typing import Callable
import warnings
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_utils.sampling_torch import (
    draw_sobol_samples,
)
from torch.quasirandom import SobolEngine


def initialize_q_batch(X: Tensor, Y: Tensor, n: int, eta=1.0):
    """Heuristic for selecting initial conditions for candidate generation.

    This heuristic selects points from 'X' (without replacement) with probability
    proportional to 'exp(eta * Z)', where 'Z = (Y - mean(Y)) / std(Y)' and 'eta'
    is a temperature parameter.

    Args:
        X: A 'b x q x d' ndarray of 'b' samples of
            'q'-batches from a 'd'-dim feature space. Typically, these are generated using qMC sampling.
        Y: A ndarray of 'b' outcomes associated with the samples.
        n: The number of initial condition to be generated. Must be less than 'b'.
        eta: Temperature parameter for weighting samples.

    Returns:
        A 'n x q x d' ndarray of 'n'-'q'-batch initial
        conditions, where each batch of 'n x q x d' samples is selected independently.
    """
    n_samples = X.shape[0]
    if n > n_samples:
        raise RuntimeError(
            f"n ({n}) cannot be larger than the number of "
            f"provided samples ({n_samples})"
        )
    elif n == n_samples:
        return X

    Ystd = Y.std(dim=0)  # shape()
    if torch.any(Ystd == 0):
        warnings.warn(
            "All acquisition values for raw samples points are the same for "
            "at least one batch. Choosing initial conditions at random."
        )
        return X[torch.randperm(n=n_samples)][
            :n
        ]  # X[np.random.permutation(n_samples)][:n]

    max_idx = torch.argmax(Y, axis=0)  # shape()
    Z = (Y - Y.mean(dim=0)) / Ystd  # shape(b)
    etaZ = eta * Z
    weights = torch.exp(etaZ)
    while torch.isinf(weights).any():
        etaZ *= 0.5
        weights = torch.exp(etaZ)
    idcs = torch.multinomial(
        weights, n
    )  # idcs = np.random.choice(len(weights), n, replace=False, p=weights)
    if max_idx not in idcs:
        idcs[-1] = max_idx
    return X[idcs]


def gen_batch_initial_conditions(
    acq_function: Callable,
    bounds: Tensor,
    q: int,
    num_restarts: int,
    raw_samples: int,
    options=None,
):
    """Generate a batch of initial conditions for random-restart optimziation.
    Args:
        acq_function: The acquisition function to be optimized.
        bounds: A '2 x d' ndarray of lower and upper bounds.
        q: The number of candidates.
        num_restarts: The number of starting points for multistart acquisition
            function optimization.
        raw_samples: The number of samples for initialization.
    Returns:
        A 'num_restart x q x d' ndarray of initial conditions.
    """
    options = options or {}
    seed = options.get("seed")
    effective_dim = bounds.shape[-1] * q

    if effective_dim > SobolEngine.MAXDIM:  # Max dimensionality of Sobol is 21201.
        warnings.warn(
            f"Sample dimension q*d={effective_dim} exceeding Sobol max dimension 21201. Using iid samples instead."
        )
        X_rnd_nlzd = torch.rand(raw_samples, q, bounds.shape[-1], dtype=bounds.dtype)
        X_rnd = bounds[0] + (bounds[1] - bounds[0]) * X_rnd_nlzd
    else:
        X_rnd = draw_sobol_samples(
            bounds=bounds, n=raw_samples, q=q, seed=seed
        )  # shape(n, q, ndims)
    ndims = bounds.shape[-1]
    X_rnd_batch = X_rnd.reshape(-1, ndims)  # shape(raw_samples * q, ndims)
    Y_rnd = acq_function(X_rnd_batch)  # shape(raw_samples * q)
    batch_initial_conditions = initialize_q_batch(X=X_rnd, Y=Y_rnd, n=num_restarts)
    return batch_initial_conditions


def gen_mll_initial_conditions(
    mll: Callable, bounds: Tensor, q: int, num_restarts: int, raw_samples: int
):
    effective_dim = bounds.shape[-1] * q

    if effective_dim > SobolEngine.MAXDIM:  # Max dimensionality of Sobol is 21201.
        warnings.warn(
            f"Sample dimension q*d={effective_dim} exceeding Sobol max dimension 21201. Using iid samples instead."
        )
        X_rnd_nlzd = torch.rand(raw_samples, q, bounds.shape[-1], dtype=bounds.dtype)
        X_rnd = bounds[0] + (bounds[1] - bounds[0]) * X_rnd_nlzd
    else:
        X_rnd = draw_sobol_samples(
            bounds=bounds, n=raw_samples, q=q
        )  # shape(n, q, ndims)
    # initialize_q_batch
    return X_rnd



import math
from tqdm import tqdm
from dataclasses import dataclass

import torch
from botorch.acquisition import qExpectedImprovement
from botorch.fit import fit_gpytorch_mll
from botorch.generation import MaxPosteriorSampling
from botorch.models import SingleTaskGP
from botorch.optim import optimize_acqf

from torch.quasirandom import SobolEngine

import gpytorch
from gpytorch.constraints import Interval
from gpytorch.kernels import MaternKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
dtype = torch.double
max_cholesky_size = float("inf")  # Always use Cholesky


@dataclass
class TurboState:
    """Maintain the TuRBO state
    TuRBO needs to maintain a state, which includes the length of the trust region,
    success and failure counters, success and failure tolerance, etc.

    In this tutorial we store the state in a dataclass and update the
    state of TuRBO after each batch evaluation.

    **Note**: These settings assume that the domain has been scaled to $[0, 1]^d$
    and that the same batch size is used for each iteration.
    """

    dim: int
    batch_size: int
    length: float = 0.8
    length_min: float = 0.5**7
    length_max: float = 1.6
    failure_counter: int = 0
    failure_tolerance: int = float("nan")  # Note: Post-initialized
    success_counter: int = 0
    success_tolerance: int = 10  # Note: The original paper uses 3
    best_value: float = -float("inf")
    restart_triggered: bool = False

    def __post_init__(self):
        self.failure_tolerance = math.ceil(
            max([4.0 / self.batch_size, float(self.dim) / self.batch_size])
        )


def update_state(state, Y_next):
    if max(Y_next) > state.best_value + 1e-3 * math.fabs(state.best_value):
        state.success_counter += 1
        state.failure_counter = 0
    else:
        state.success_counter = 0
        state.failure_counter += 1

    if state.success_counter == state.success_tolerance:  # Expand trust region
        state.length = min(2.0 * state.length, state.length_max)
        state.success_counter = 0
    elif state.failure_counter == state.failure_tolerance:  # Shrink trust region
        state.length /= 2.0
        state.failure_counter = 0

    state.best_value = max(state.best_value, max(Y_next).item())
    if state.length < state.length_min:
        state.restart_triggered = True
    return state


def get_initial_points(dim, n_pts, seed=0):
    """Generate initial points
    This generates an initial set of Sobol points that we use to start of the BO loop.
    """
    sobol = SobolEngine(dimension=dim, scramble=True, seed=seed)
    X_init = sobol.draw(n=n_pts).to(dtype=dtype, device=device)
    return X_init


def generate_batch(
    state,
    model,  # GP model
    X,  # Evaluated points on the domain [0, 1]^d
    Y,  # Function values
    batch_size,
    n_candidates=None,  # Number of candidates for Thompson sampling
    num_restarts=10,
    raw_samples=512,
    acqf="ts",  # "ei" or "ts"
):
    """Generate new batch
    Given the current `state` and a probabilistic (GP) `model` built from observations `X` and `Y`,
    we generate a new batch of points.

    This method works on the domain $[0, 1]^d$, so make sure to not pass in observations from the true domain.
    `unnormalize` is called before the true function is evaluated which will first map the points back to the original domain.

    We support either TS and qEI which can be specified via the `acqf` argument.
    """
    assert acqf in ("ts", "ei")
    assert X.min() >= 0.0 and X.max() <= 1.0 and torch.all(torch.isfinite(Y))
    if n_candidates is None:
        n_candidates = min(5000, max(2000, 200 * X.shape[-1]))

    # Scale the TR to be proportional to the lengthscales
    x_center = X[Y.argmax(), :].clone()
    weights = model.covar_module.base_kernel.lengthscale.squeeze().detach()
    weights = weights / weights.mean()
    weights = weights / torch.prod(weights.pow(1.0 / len(weights)))
    tr_lb = torch.clamp(x_center - weights * state.length / 2.0, 0.0, 1.0)
    tr_ub = torch.clamp(x_center + weights * state.length / 2.0, 0.0, 1.0)

    if acqf == "ts":
        dim = X.shape[-1]
        sobol = SobolEngine(dim, scramble=True)
        pert = sobol.draw(n_candidates).to(dtype=dtype, device=device)
        pert = tr_lb + (tr_ub - tr_lb) * pert

        # Create a perturbation mask
        prob_perturb = min(20.0 / dim, 1.0)
        mask = torch.rand(n_candidates, dim, dtype=dtype, device=device) <= prob_perturb
        ind = torch.where(mask.sum(dim=1) == 0)[0]
        mask[ind, torch.randint(0, dim - 1, size=(len(ind),), device=device)] = 1

        # Create candidate points from the perturbations and the mask
        X_cand = x_center.expand(n_candidates, dim).clone()
        X_cand[mask] = pert[mask]

        # Sample on the candidate points
        thompson_sampling = MaxPosteriorSampling(model=model, replacement=False)
        with torch.no_grad():  # We don't need gradients when using TS
            X_next = thompson_sampling(X_cand, num_samples=batch_size)

    elif acqf == "ei":
        ei = qExpectedImprovement(model, train_Y.max(), maximize=True)
        X_next, acq_value = optimize_acqf(
            ei,
            bounds=torch.stack([tr_lb, tr_ub]),
            q=batch_size,
            num_restarts=num_restarts,
            raw_samples=raw_samples,
        )

    return X_next


def turbo(eval_func, dim, n_init, n_iterations, batch_size=1):
    X_turbo = get_initial_points(dim, n_init)  # [n_init, dim]
    Y_turbo = torch.tensor(
        [eval_func(x) for x in X_turbo], dtype=dtype, device=device
    ).unsqueeze(-1)

    state = TurboState(dim, batch_size=batch_size)

    NUM_RESTARTS = 10
    RAW_SAMPLES = 512
    N_CANDIDATES = min(5000, max(2000, 200 * dim))

    torch.manual_seed(0)
    c = 0
    pbar = tqdm(total=n_iterations)
    while not state.restart_triggered and c < n_iterations:  # Run until TuRBO converges
        c += 1
        # Fit a GP model
        train_Y = (Y_turbo - Y_turbo.mean()) / Y_turbo.std()
        likelihood = GaussianLikelihood(noise_constraint=Interval(1e-8, 1e-3))
        covar_module = (
            ScaleKernel(  # Use the same lengthscale prior as in the TuRBO paper
                MaternKernel(
                    nu=2.5,
                    ard_num_dims=dim,
                    lengthscale_constraint=Interval(0.005, 4.0),
                )
            )
        )
        model = SingleTaskGP(
            X_turbo, train_Y, covar_module=covar_module, likelihood=likelihood
        )
        mll = ExactMarginalLogLikelihood(model.likelihood, model)

        # Do the fitting and acquisition function optimization inside the Cholesky context
        with gpytorch.settings.max_cholesky_size(max_cholesky_size):
            # Fit the model
            fit_gpytorch_mll(mll)

            # Create a batch
            X_next = generate_batch(
                state=state,
                model=model,
                X=X_turbo,
                Y=train_Y,
                batch_size=batch_size,
                n_candidates=N_CANDIDATES,
                num_restarts=NUM_RESTARTS,
                raw_samples=RAW_SAMPLES,
                acqf="ts",
            )

        Y_next = torch.tensor(
            [eval_func(x) for x in X_next], dtype=dtype, device=device
        ).unsqueeze(-1)

        # Update state
        state = update_state(state=state, Y_next=Y_next)

        # Append data
        X_turbo = torch.cat((X_turbo, X_next), dim=0)
        Y_turbo = torch.cat((Y_turbo, Y_next), dim=0)

        # Print current status
        pbar.update()
        pbar.set_postfix_str(
            f"{len(X_turbo)}) Best value: {state.best_value:.2e}, TR length: {state.length:.2e}"
        )
    return X_turbo, Y_turbo


if __name__ == "__main__":
    from botorch.test_functions import Ackley

    fun = Ackley(dim=20, negate=True).to(dtype=dtype, device=device)
    fun.bounds[0, :].fill_(-5)
    fun.bounds[1, :].fill_(10)
    DIM = fun.dim

    BATCH_SIZE = 4
    N_INIT = 2 * DIM

    def eval_objective(x):
        """This is a helper function we use to unnormalize and evalaute a point"""
        lb, ub = fun.bounds
        return fun(lb + (ub - lb) * x)

    X, Y = turbo(
        eval_func=eval_objective,
        dim=DIM,
        n_init=N_INIT,
        n_iterations=8,
        batch_size=BATCH_SIZE,
    )
    import numpy as np
    import matplotlib.pyplot as plt

    fx = np.maximum.accumulate(Y.cpu())
    plt.plot(fx, marker="", lw=3)

    plt.plot([0, len(Y)], [fun.optimal_value, fun.optimal_value], "k--", lw=3)
    plt.show()



from .Ackley import Ackley
from .Bukin import Bukin
from .Branin import Branin
from .Dropwave import DropWave
from .Eggholder import EggHolder
from .Griewank import Griewank
from .Hartmann6d import Hartmann6d
from .Holdertable import HolderTable
from .Productsines import ProductSines
from .Rastrigin import Rastrigin
from .Rosenbrock import Rosenbrock
from .Shubert import Shubert
from .Sphere import Sphere
from .Trid import Trid
from .Noise import Noise
from .Zero import Zero
from .Sin import Sin
from .AbsMax import AbsMax

Base_func_list = [
    Ackley,
    Bukin,
    Branin,
    DropWave,
    EggHolder,
    Griewank,
    Hartmann6d,
    HolderTable,
    ProductSines,
    Rastrigin,
    Rosenbrock,
    Shubert,
    Sphere,
    Trid,
    Noise,
    Zero,
    Sin,
    AbsMax,
]



import numpy as np
from scipy.stats.qmc import Sobol
from math import log2


class SobolSampler:
    def __init__(self, dim, scramble=True, seed=None):
        self.seed_generator = np.random.default_rng(seed)
        self.get_sampler = lambda: Sobol(
            d=dim,
            scramble=scramble,
            # optimization="random-cd",  # extremely slow
            seed=self.seed_generator.integers(10000),
        )

    def sample(self, n):
        sampler = self.get_sampler()
        if n & (n - 1) == 0:  # There exists integer m such that n=2^m.
            m = int(log2(n))
            x = sampler.random_base2(m=m)
        else:
            x = sampler.random(n)
        x = x * 2 - 1
        return x



import torch
from torch import Tensor
from torch.linalg import qr
from .normRotMargLikelihood import norm_rot_marg_likelihood
from HDBOBenchmark.algorithms.add_bo_quasi_nt.utils.orthToPermutation import (
    orth_to_permutation,
)
from .decompOptPartial import decomp_opt_partial
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_optim.optimize import optimize_mll


def add_gp_decomp_marg_likelihood(x: Tensor, y: Tensor, decomp, hyper_params):
    """
    This function attempts to find the best Kernel hyper parameters to fit an additive function.
    The kernel parameters include the smoothness and scale paramters and the decomposition.
    Args:
        x: Training data
        y: Training target
        decomp: If the decomposition need not be learned, then this should contain the true decomposition.
            Otherwise, it should contain two fields d (# of dimensions per group) and M (number of groups).
        hyper_params: should have a field called decompStrategy: It should be one of 'known', 'learn', 'random' and 'partialLearn'.
            'known': The decomposition is known and given in decomp. We will optimize according to this.
            'learn': The decomposition is unknown and should be learned.
            'random': Randomly pick a partition at each iteration.
            'partialLearn': Partially learn the decomposition at each iteration by trying out a few and picking the best.
    Returns:
    """
    # Prelims
    ndims = x.shape[1]

    if hyper_params.decomp_strategy != "known":
        if hasattr(decomp, "m"):
            # First create a placeholder for the decomposition
            raise NotImplementedError
        else:  # then decomp is a vector of values with the number of dims in each group.
            m, p = len(decomp), torch.sum(decomp)
            cum_dims = torch.cumsum(decomp, dim=0)
            cum_dims = torch.hstack((torch.IntTensor([0]), cum_dims))
            decomposition = []
            for i in range(m):
                decomposition.append(torch.arange(cum_dims[i], cum_dims[i + 1]))
    else:
        decomposition = decomp
    num_groups = len(decomposition)
    one_vec = torch.ones(num_groups).to(x)

    # Set the Hyperparameters for each GP
    # Common Mean Function
    common_mean_func = hyper_params.common_mean_func
    # Common Noise parameter
    common_noise = hyper_params.common_noise
    # Mean Function for each GP
    mean_funcs = [hyper_params.mean_funcs] * num_groups
    # Noise Parameters
    noises = hyper_params.noises
    # Some parameters for learning the Kernel smoothness and range
    # Define Bounds for optimization of bandwidth and scale
    sigma_sm_bound = hyper_params.sigma_sm_range  # bandwidth bounds
    sigma_pr_bound = hyper_params.sigma_pr_range  # scale bounds
    bounds = torch.log(torch.vstack((sigma_sm_bound, sigma_pr_bound)))

    # Learn the Decomposition
    if hyper_params.decomp_strategy != "known":
        a = torch.randn((ndims, p), dtype=x.dtype)
        a, _ = qr(a)
        # a = torch.eye(ndims, dtype=x.dtype)
        # maximize marginal likelihood
        # Using shgo
        """
        res = shgo(lambda t: -norm_rot_marg_likelihood(np.exp(t[0]) * one_vec, np.exp(t[1]) * one_vec, decomposition, a, x, y, mean_funcs, common_mean_func, noises, common_noise),
                   bounds)
        solution = res.x
        """

        # Using quasi-Newton method
        def mll(t):
            return norm_rot_marg_likelihood(
                torch.exp(t[0]) * one_vec,
                torch.exp(t[1]) * one_vec,
                decomposition,
                a,
                x,
                y,
                mean_funcs,
                common_mean_func,
                noises,
                common_noise,
            )

        opt_params, _ = optimize_mll(
            mll=mll, bounds=bounds.T, q=1, num_restarts=128, raw_samples=128
        )
        opt_params = opt_params.squeeze()

        sigma_sm_opts = torch.exp(opt_params[0]) * one_vec
        sigma_pr_opts = torch.exp(opt_params[1]) * one_vec

        # Now optimize w.r.t a
        if hyper_params.decomp_strategy == "random":
            raise NotImplementedError
        elif hyper_params.decomp_strategy == "partialLearn":
            a = decomp_opt_partial(
                lambda t: -norm_rot_marg_likelihood(
                    sigma_sm_opts,
                    sigma_pr_opts,
                    decomposition,
                    t,
                    x,
                    y,
                    mean_funcs,
                    common_mean_func,
                    noises,
                    common_noise,
                ),
                ndims,
            )
        elif hyper_params.decomp_strategy == "learn":
            raise NotImplementedError
        else:
            raise Exception("Unknown Strategy to handle decomposition\n")
    else:
        # Otherwise, use the given decomposition
        a = torch.eye(ndims).to(x)

    # Learn the Decomposition Ends here

    # Finally, optimize w.r.t sigma and scale again.
    def mll(t):
        return norm_rot_marg_likelihood(
            torch.exp(t[0]) * one_vec,
            torch.exp(t[1]) * one_vec,
            decomposition,
            a,
            x,
            y,
            mean_funcs,
            common_mean_func,
            noises,
            common_noise,
        )

    """
    res = shgo(lambda t: -norm_rot_marg_likelihood(np.exp(t[0]) * one_vec, np.exp(t[1]) * one_vec, decomposition, a, x, y, mean_funcs, common_mean_func, noises, common_noise),
               bounds)
    marg_like_val, opt_params = -res.fun, res.x
    """
    # Using quasi-Newton method
    opt_params, marg_like_val = optimize_mll(
        mll=mll, bounds=bounds.T, q=1, num_restarts=128, raw_samples=128
    )
    opt_params = opt_params.squeeze()

    sigma_sm_opts = torch.exp(opt_params[0]) * one_vec
    sigma_pr_opts = torch.exp(opt_params[1]) * one_vec

    # Finally, return the learned decomposition
    learned_decomp = []
    if hyper_params.decomp_strategy == "partialLearn":
        permute_order = torch.arange(ndims, dtype=torch.long) @ a.long()
        for i in range(num_groups):
            learned_decomp.append(permute_order[decomposition[i]])
    elif hyper_params.decomp_strategy == "random":
        row_ind, col_ind = orth_to_permutation(a.numpy())
        row_ind, col_ind = torch.from_numpy(row_ind), torch.from_numpy(col_ind)
        indices = torch.argsort(col_ind)
        row_ind = row_ind[indices]
        learned_decomp = []
        for i in range(num_groups):
            learned_decomp.append(row_ind[decomposition[i]])
    elif hyper_params.decomp_strategy == "known":
        learned_decomp = decomposition
    else:
        raise NotImplementedError
    """
    row_ind, col_ind = orth_to_permutation(a.numpy())
    row_ind, col_ind = torch.from_numpy(row_ind), torch.from_numpy(col_ind)
    indices = torch.argsort(col_ind)
    row_ind = row_ind[indices]
    learned_decomp = []
    for i in range(num_groups):
        learned_decomp.append(row_ind[decomposition[i]])
    """
    return sigma_sm_opts, sigma_pr_opts, a, learned_decomp, marg_like_val



import numpy as np
from .param import Parameter


class StepIntPara(Parameter):
    """
    Integer parameter, that increments with a fixed step, like `[4, 8, 12, 16]`
    The config would be like `{'name' : 'x', 'type' : 'step_int', 'lb' : 4, 'ub' : 16, 'step' : 4}`
    """

    def __init__(self, param_dict):
        super().__init__(param_dict)
        self.lb = round(param_dict["lb"])
        self.ub = round(param_dict["ub"])
        self.step = round(param_dict["step"])
        self.num_step = (self.ub - self.lb) // self.step

    def sample(self, num=1):
        return np.random.randint(0, self.num_step + 1, num) * self.step + self.lb

    def transform(self, x: np.ndarray) -> np.ndarray:
        return (x - self.lb) / self.step

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        x_recover = x * self.step + self.lb
        return x_recover.round().astype(int)

    @property
    def is_numeric(self):
        return True

    @property
    def opt_lb(self):
        return 0.0

    @property
    def opt_ub(self):
        return 1.0 * self.num_step

    @property
    def is_discrete(self):
        return True

    @property
    def is_discrete_after_transform(self):
        return True



import logging
from logging import Logger as _Logger
import time


class Logger(_Logger):
    def __init__(self, name, level=0):
        super().__init__(name, level)
        self.start_time = time.time()

    def _process(self, msg):
        return f"[{time_text(time.time() - self.start_time)}] " + msg

    def callHandlers(self, record):
        if not "msg_type" in record.__dict__:
            record.msg = self._process(record.msg)
        super().callHandlers(record)


logging.basicConfig(level="INFO", format="%(message)s")
logging.setLoggerClass(Logger)
logger = logging.getLogger(__name__)


def time_dhms(s):
    s = int(s)
    day = s // 86400
    s %= 86400
    hour = s // 3600
    s %= 3600
    minute = s // 60
    s %= 60
    return day, hour, minute, s


def time_text(s):
    d, h, m, s = time_dhms(s)
    return f"{d}d {h:>2d}:{m:>2d}:{s:>2d}"



from .template import Args as ArgsTemplate
from HDBOBenchmark.funcs.base.Holdertable import HolderTable as get_func


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = "HolderTable"
        # BO
        self.dims = 2
        self.get_func = get_func
        self.func = get_func(self.dims)  # self.n_group)

        self.method_kwargs["hebo"]["model_name"] = "gpy"


args = Args()



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy as np
from .param import Parameter


class CategoricalPara(Parameter):
    def __init__(self, param):
        super().__init__(param)
        self.categories = list(param["categories"])
        try:
            self._categories_dict = {k: v for v, k in enumerate(self.categories)}
        except TypeError:  # there are unhashable types
            self._categories_dict = None
        self.lb = 0
        self.ub = len(self.categories) - 1

    def sample(self, num=1):
        if num <= 0:
            raise ValueError(f"num <= 0: {num}")
        return np.random.choice(self.categories, num, replace=True)

    def transform(self, x: np.ndarray):
        if self._categories_dict:
            # if all objects are hashable, we can use a dict instead for faster transform
            ret = np.array(list(map(lambda a: self._categories_dict[a], x)))
        else:
            # otherwise, we fall back to searching in an array
            ret = np.array(list(map(lambda a: np.where(self.categories == a)[0][0], x)))
        return ret.astype(float)

    def inverse_transform(self, x):
        return np.array([self.categories[x_] for x_ in x.round().astype(int)])

    @property
    def is_numeric(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_discrete_after_transform(self):
        return True

    @property
    def opt_lb(self):
        return self.lb

    @property
    def opt_ub(self):
        return self.ub



from HDBOBenchmark.args.base.template import Args as ArgsTemplate
from HDBOBenchmark.funcs.realistic.MIPbenchmark.MpsModel import MPSModel as get_func

# enlight_hard sp150x300d neos-787933 nu25-pr12 neos-1122047 neos-1171448 neos-827175


class SCIPArgs(ArgsTemplate):
    def __init__(self, mps_path="enlight_hard", t_max=20):
        super().__init__()
        self.name = "miplib"
        # BO
        self.dims = 135
        self.get_func = get_func
        # enlight_hard,
        self.func = get_func(
            mps_path=f"revised-submissions/miplib2010_publically_available/instances/{mps_path}.mps.gz",
            solu_path="miplib2017-v26.solu",
            solver="scip",
            t_max=t_max,  # upto 2 hours (7200)
            auto_init_y_scaler=False,
        )

        self.method_kwargs["hebo"][
            "model_name"
        ] = "gumbel"  # gp for dim 5, gumbel for dim 10 20


class RawArgs(ArgsTemplate):
    def __init__(self, mps_path="enlight_hard"):
        super().__init__()
        self.name = "miplib"
        # BO
        self.dims = 34
        self.get_func = get_func
        # enlight_hard,
        self.func = get_func(
            mps_path=f"revised-submissions/miplib2010_publically_available/instances/{mps_path}.mps.gz",
            solu_path="miplib2017-v26.solu",
            solver=None,
        )

        self.method_kwargs["hebo"][
            "model_name"
        ] = "gumbel"  # gp for dim 5, gumbel for dim 10 20


args = SCIPArgs(mps_path="enlight_hard")



import torch
from torch import Tensor
from torch.quasirandom import SobolEngine
from typing import Optional


def draw_sobol_samples(
    bounds: Tensor,
    n: int,
    q: int,
    batch_shape: Optional[torch.Size] = None,
    seed: Optional[int] = None,
) -> Tensor:
    r"""Draw qMC samples from the box defined by bounds.

    Args:
        bounds: A `2 x d` dimensional tensor specifying box constraints on a
            `d`-dimensional space, where bounds[0, :] and bounds[1, :] correspond
            to lower and upper bounds, respectively.
        n: The number of (q-batch) samples. As a best practice, use powers of 2.
        q: The size of each q-batch.
        batch_shape: The batch shape of the samples. If given, returns samples
            of shape `n x batch_shape x q x d`, where each batch is an
            `n x q x d`-dim tensor of qMC samples.
        seed: The seed used for initializing Owen scrambling. If None (default),
            use a random seed.

    Returns:
        A `n x batch_shape x q x d`-dim tensor of qMC samples from the box
        defined by bounds.

    Example:
        >>> bounds = torch.stack([torch.zeros(3), torch.ones(3)])
        >>> samples = draw_sobol_samples(bounds, 16, 2)
    """
    batch_shape = batch_shape or torch.Size()
    batch_size = int(torch.prod(torch.tensor(batch_shape)))
    d = bounds.shape[-1]
    lower = bounds[0]
    rng = bounds[1] - bounds[0]
    sobol_engine = SobolEngine(q * d, scramble=True, seed=seed)
    samples_raw = sobol_engine.draw(batch_size * n, dtype=lower.dtype)
    samples_raw = samples_raw.view(*batch_shape, n, q, d).to(device=lower.device)
    if batch_shape != torch.Size():
        samples_raw = samples_raw.permute(-3, *range(len(batch_shape)), -2, -1)
    return lower + rng * samples_raw


def draw_sobol_samples_(
    bounds: Tensor,
    n: int,
    q: int,
    batch_shape: Optional[torch.Size] = None,
    seed: Optional[int] = None,
) -> Tensor:
    r"""Draw qMC samples from the box defined by bounds.

    Args:
        bounds: A `2 x d` dimensional tensor specifying box constraints on a
            `d`-dimensional space, where bounds[0, :] and bounds[1, :] correspond
            to lower and upper bounds, respectively.
        n: The number of (q-batch) samples. As a best practice, use powers of 2.
        q: The size of each q-batch.
        batch_shape: The batch shape of the samples. If given, returns samples
            of shape `n x batch_shape x q x d`, where each batch is an
            `n x q x d`-dim tensor of qMC samples.
        seed: The seed used for initializing Owen scrambling. If None (default),
            use a random seed.

    Returns:
        A `n x batch_shape x q x d`-dim tensor of qMC samples from the box
        defined by bounds.

    Example:
        >>> bounds = torch.stack([torch.zeros(3), torch.ones(3)])
        >>> samples = draw_sobol_samples(bounds, 16, 2)
    """
    batch_shape = batch_shape or torch.Size()
    batch_size = int(torch.prod(torch.tensor(batch_shape)))
    d = bounds.shape[-1]
    lower = bounds[0]
    rng = bounds[1] - bounds[0]
    sobol_engine = SobolEngine(1, scramble=True, seed=seed)
    samples_raw = sobol_engine.draw(batch_size * n, dtype=lower.dtype)
    samples_raw = torch.repeat_interleave(samples_raw, q * d, dim=1)
    samples_raw = samples_raw.view(*batch_shape, n, q, d).to(device=lower.device)
    if batch_shape != torch.Size():
        samples_raw = samples_raw.permute(-3, *range(len(batch_shape)), -2, -1)
    return lower + rng * samples_raw



import numpy as np


def project_to_rectangle(pt, bounds=None):
    """
    :param pt: (dims,)
    :param bounds: (dims, 2) ndarray with the lower bounds on the first column and the upper bound on the second column.
        If bounds is not given, it will be taken as [0,1]^d.
    :return:
    """
    dims = pt.shape[0]
    if bounds is None:
        bounds = np.array([[0, 1] * dims])
    below_lbs = (pt - bounds[:, 0]) < 0
    pt[below_lbs] = bounds[below_lbs, 0]

    above_ubs = (pt - bounds[:, 1]) > 0
    pt[above_ubs] = bounds[above_ubs, 1]
    return pt



from math import log2
from scipy.stats.qmc import Sobol


def draw_sobol_samples(bounds, n, q, seed=None):
    """Draw qMC samples from the box defined by bounds.

    Args:
        bounds: A '2 x d' dimensional ndarray specifying box constraints on a
            'd'-dimensional space, where bounds[0, :] and bounds[1, :] correspond
            to lower and upper bounds, respectively.
        n: The number of (q-batch) samples. As a best practice, use powers of 2.
        q: The size of each q-batch.
        seed: The seed used for initializing Owen scrambling. If None (default),
            use a random seed.

    Returns:
        A 'n x q x d'-dimensional ndarray of qMC samples from the box
        defined by bounds.
    """
    ndims = bounds.shape[-1]
    lower = bounds[0]
    rng = bounds[1] - bounds[0]
    sampler = Sobol(d=q * ndims, scramble=True)
    if n & (n - 1) == 0:  # There exists integer m such that n=2^m.
        m = int(log2(n))
        samples_raw = sampler.random_base2(m)
    else:
        samples_raw = sampler.random(n)
    samples_raw = samples_raw.reshape((n, q, ndims))
    return lower + rng * samples_raw


def draw_sobol_samples_common(bounds, n):
    """Draw qMC samples from the box defined by bounds.

    Args:
        bounds: A '2 x d' dimensional ndarray specifying box constraints on a
            'd'-dimensional space, where bounds[0, :] and bounds[1, :] correspond
            to lower and upper bounds, respectively.
        n: The number of (q-batch) samples. As a best practice, use powers of 2.

    Returns:
        A 'n x d'-dimensional ndarray of qMC samples from the box
        defined by bounds.
    """
    ndims = bounds.shape[-1]
    lower = bounds[0]
    rng = bounds[1] - bounds[0]
    sampler = Sobol(d=ndims, scramble=True)
    if n & (n - 1) == 0:  # There exists integer m such that n=2^m.
        m = int(log2(n))
        samples_raw = sampler.random_base2(m)
    else:
        samples_raw = sampler.random(n)
    return lower + rng * samples_raw



import pickle
import os
import re


def save_pickle(path, *data):
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(os.path.dirname(path))
    with open(path, "wb") as f:
        pickle.dump(data, f, True)


def load_pickle(path):
    with open(path, "rb") as f:
        data = pickle.load(f)
    return data


def path_find_file(path, pattern):
    files_name = os.listdir(path)
    for name in files_name:
        file = os.path.join(path, name)
        if os.path.isfile(file) and (re.match(pattern, name) is not None):
            return name
    return None



import torch

"""
def get_rand_perm_mat(dims):
    p = np.zeros((dims, dims))
    shuffle_order = np.random.permutation(dims)
    for i in range(dims):
        p[i, shuffle_order[i]] = 1
    return p
"""


def get_rand_perm_mat(ndims: int):
    p = torch.zeros(ndims, ndims, dtype=torch.double)
    shuffle_order = torch.randperm(ndims)
    for i in range(ndims):
        p[i, shuffle_order[i]] = 1
    return p



from .template import Args as ArgsTemplate
from HDBOBenchmark.funcs.base.BBOB import bbob_func_dict

func_name = "Rastrigin"


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = func_name
        # BO
        self.dims = 10
        self.func = bbob_func_dict[func_name]

        self.method_kwargs["hebo"][
            "model_name"
        ] = "gp"  # gp for dim 5, gumbel for dim 10 20


args = Args()



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

from . import design_space
from . import param
from . import bool_param
from . import categorical_param
from . import integer_param
from . import numeric_param
from . import pow_param
from . import pow_integer_param
from . import step_int
from .design_space import DesignSpace



def un_gz(file_name: str, dest: str) -> str:
    import gzip

    """unzip .gz file

    Args:
        file_name (str): The file to unzip
        dest (str): unzip destination path/file name

    Returns:
        str: destination path/file name
    """

    f_name = file_name.replace(".gz", "")
    print(f"current file: {f_name}")
    g_file = gzip.GzipFile(file_name)
    if dest is not None:
        path = dest
    else:
        path = f_name
    open(path, "wb+").write(g_file.read())
    return path


def determine_file_type(file_name: str, file_type: str) -> bool:
    return file_name.split(".")[-1] == file_type


def is_gz(file_name) -> bool:
    return determine_file_type(file_name, "gz")


def is_mps(file_name) -> bool:
    return determine_file_type(file_name, "mps")


def is_npy(file_name) -> bool:
    return determine_file_type(file_name, "npy")


def get_keys_from_dict(dict, keys):
    from operator import itemgetter

    values = itemgetter(*keys)(dict)
    return values


def get_solutions(file_path="assets/miplib/miplib2017-v26.solu") -> dict:
    import math

    UNKNOWN = "=unkn="
    UNBOUND = "=unbd="
    INF = "=inf="

    with open(file_path, "r") as f:
        solutions = f.readlines()
        results = {}
        for i, solution in enumerate(solutions):
            solution_list = solution.split()
            result = {
                "type": solution_list[0],
                "filename": solution_list[1],
            }
            if result["type"] in [UNKNOWN, UNBOUND]:
                result["answer"] = None
            elif result["type"] == INF:
                result["answer"] = math.inf
            else:
                result["answer"] = solution_list[2]
            results[result["filename"]] = result

    return results


def download_miplib(path=r"assets/miplib/", target="easy-v11.test"):
    from pathlib import Path
    import requests

    base_dir = Path(path)
    base_dir.mkdir(parents=True, exist_ok=True)
    target_file = base_dir / target
    if not target_file.exists():
        print(f"Downloading {target} into {path}.")
        url = "https://miplib.zib.de/downloads/"
        r = requests.get(url + target)
        r.raise_for_status()
        with target_file.open("wb") as f:
            f.write(r.content)
    if ".zip" in target:
        import zipfile

        with zipfile.ZipFile(target_file.resolve(), mode="r") as archive:
            archive.extractall(path)



from tqdm import tqdm
import torch
from torch.quasirandom import SobolEngine

from botorch import fit_fully_bayesian_model_nuts
from botorch.acquisition import qExpectedImprovement
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.transforms import Standardize
from botorch.optim import optimize_acqf

WARMUP_STEPS = 256
NUM_SAMPLES = 128
THINNING = 16

tkwargs = {
    "device": torch.device("cuda:1" if torch.cuda.is_available() else "cpu"),
    "dtype": torch.double,
}


def saasbo(eval_func, ndims, n_iterations, n_init=10, batch_size=1):
    X = SobolEngine(dimension=ndims, scramble=True, seed=0).draw(n_init).to(**tkwargs)
    Y = torch.tensor([eval_func(x) for x in X]).unsqueeze(-1).to(**tkwargs)

    pbar = tqdm(total=n_iterations)
    for i in range(n_iterations):
        train_Y = (Y - Y.mean()) / Y.std()
        gp = SaasFullyBayesianSingleTaskGP(
            train_X=X,
            train_Y=train_Y,
            train_Yvar=torch.full_like(train_Y, 1e-6),
            outcome_transform=Standardize(m=1),
        )
        fit_fully_bayesian_model_nuts(
            gp,
            warmup_steps=WARMUP_STEPS,
            num_samples=NUM_SAMPLES,
            thinning=THINNING,
            disable_progbar=True,
        )

        EI = qExpectedImprovement(model=gp, best_f=train_Y.max())
        candidates, acq_values = optimize_acqf(
            EI,
            bounds=torch.cat((torch.zeros(1, ndims), torch.ones(1, ndims))).to(
                **tkwargs
            ),
            q=batch_size,
            num_restarts=10,
            raw_samples=1024,
        )

        Y_next = torch.cat([eval_func(x).unsqueeze(-1) for x in candidates]).unsqueeze(
            -1
        )
        X = torch.cat((X, candidates))
        Y = torch.cat((Y, Y_next))
        pbar.update()

    return X, Y


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    from botorch.test_functions import Branin

    branin = Branin().to(**tkwargs)

    def branin_emb(x):
        """x is assumed to be in [0, 1]^d"""
        lb, ub = branin.bounds
        return (
            branin(lb + (ub - lb) * x[..., :2]) * -1
        )  # Flip the value for minimization

    DIM = 30
    X, Y = saasbo(
        eval_func=branin_emb, ndims=DIM, n_iterations=8, n_init=10, batch_size=5
    )
    Y_np = -1 * Y.cpu().numpy()
    fig, ax = plt.subplots(figsize=(8, 6))
    ax.plot(np.minimum.accumulate(Y_np), color="b", label="SAASBO")
    ax.plot([0, len(Y_np)], [0.398, 0.398], "--", c="g", lw=3, label="Optimal value")
    ax.grid(True)
    ax.set_title(f"Branin, D = {DIM}", fontsize=20)
    ax.set_xlabel("Number of evaluations", fontsize=20)
    ax.set_xlim([0, len(Y_np)])
    ax.set_ylabel("Best value found", fontsize=20)
    ax.set_ylim([0, 8])
    ax.legend(fontsize=18)
    plt.savefig("results.png")



from .template import Args as ArgsTemplate
from HDBOBenchmark.funcs.base.Griewank import Griewank as get_func


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = "Griewank"
        # BO
        self.dims = 5
        self.get_func = get_func
        self.func = get_func(self.dims)  # self.n_group)

        self.method_kwargs["hebo"][
            "model_name"
        ] = "gp"  # gp for dim 5, gumbel fe for dim 10 20


args = Args()



import numpy as np
import torch
from torch import Tensor
from typing import Optional, Union
from scipy.optimize import Bounds


def _arrayify(X: Tensor) -> np.ndarray:
    r"""Convert a torch.Tensor (any dtype or device) to a numpy (double) array.

    Args:
        X: The input tensor.

    Returns:
        A numpy array of double dtype with the same shape and data as `X`.
    """
    return X.cpu().detach().contiguous().double().clone().numpy()


def make_scipy_bounds(
    X: Tensor,
    lower_bounds: Optional[Union[float, Tensor]] = None,
    upper_bounds: Optional[Union[float, Tensor]] = None,
) -> Optional[Bounds]:
    r"""Creates a scipy Bounds object for optimziation

    Args:
        X: `... x d` tensor
        lower_bounds: Lower bounds on each column (last dimension) of `X`. If
            this is a single float, then all columns have the same bound.
        upper_bounds: Lower bounds on each column (last dimension) of `X`. If
            this is a single float, then all columns have the same bound.

    Returns:
        A scipy `Bounds` object if either lower_bounds or upper_bounds is not
        None, and None otherwise.

    Example:
        >>> X = torch.rand(5, 2)
        >>> scipy_bounds = make_scipy_bounds(X, 0.1, 0.8)
    """
    if lower_bounds is None and upper_bounds is None:
        return None

    def _expand(bounds: Union[float, Tensor], X: Tensor, lower: bool) -> Tensor:
        if bounds is None:
            ebounds = torch.full_like(X, float("-inf" if lower else "inf"))
        else:
            if not torch.is_tensor(bounds):
                bounds = torch.tensor(bounds)
            ebounds = bounds.expand_as(X)
        return _arrayify(ebounds).flatten()

    lb = _expand(bounds=lower_bounds, X=X, lower=True)
    ub = _expand(bounds=upper_bounds, X=X, lower=False)
    return Bounds(lb=lb, ub=ub, keep_feasible=True)


def np_make_scipy_bounds(X, lower_bounds, upper_bounds):
    """Creates a scipy Bounds object for optimization

    Args:
        X: '... x d' ndarray
        lower_bounds: Lower bounds on each column (last dimension) of `X`. If
            this is a single float, then all columns have the same bound.
        upper_bounds: Lower bounds on each column (last dimension) of `X`. If
            this is a single float, then all columns have the same bound.
    Returns:
        A scipy 'Bounds' object if either lower_bounds or upper_bounds is not
        None, and None otherwise.
    """
    if lower_bounds is None and upper_bounds is None:
        return None

    if lower_bounds is None:
        lb = np.full_like(X, -np.inf)
    else:
        lb = np.broadcast_to(lower_bounds, X.shape)
    if upper_bounds is None:
        ub = np.full_like(X, np.inf)
    else:
        ub = np.broadcast_to(upper_bounds, X.shape)
    return Bounds(lb.flatten(), ub.flatten(), keep_feasible=True)



from .template import Args as ArgsTemplate
from HDBOBenchmark.funcs.base.Shubert import Shubert as get_func


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = "Shubert"
        # BO
        self.dims = 10
        self.get_func = get_func
        self.func = get_func(self.dims)

        self.method_kwargs["hebo"]["model_name"] = "gpy"  # gpy for dim 5 10


args = Args()



"""
==================================================
Repo: https://github.com/ISosnovik/top
==================================================
"""
import numpy as np
from pathlib import Path
from typing import Optional
from HDBOBenchmark.funcs.base.FunctionBase import TestFunction


class Topology(TestFunction):
    def __init__(
        self,
        dim=1600,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=False,
        data_dir="assets/topology/",
    ):
        self.load_data(data_dir)
        super().__init__(1600, "Topology")

        self._x_bound = np.array([[0, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        self.y_scale = 2
        self._min_pt = None
        self._min_val = -1

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)
        self.switch_maximum()

    def load_data(self, data_dir="assets/topology/"):
        data_dir = Path(data_dir)
        if not data_dir.exists():
            data_dir.mkdir(parents=True, exist_ok=True)
        cur_data = data_dir / "target_bin.npy"

        if not cur_data.exists():
            print(f"Downloading {str(cur_data)}")
            import requests

            url_dir = "https://raw.githubusercontent.com/Yiyuiii/HDBO-B/master/assets/topology/"
            r = requests.get(url_dir + "target_bin.npy")
            r.raise_for_status()
            with cur_data.open("wb") as f:
                f.write(r.content)
        self.target = np.load(str(cur_data)).flatten()

    @staticmethod
    def score_function(predicted, target, metric: Optional[str] = "cos"):
        if not isinstance(predicted, np.ndarray):
            predicted = np.asarray(predicted).flatten()
        if not isinstance(target, np.ndarray):
            target = np.asarray(target)

        if target.ndim < predicted.ndim:
            target = np.expand_dims(target, axis=0)
        assert target.ndim == predicted.ndim

        if metric == "cos":
            num = float(np.dot(predicted, target))  # 向量点乘
            denom = np.linalg.norm(predicted, axis=-1) * np.linalg.norm(target, axis=-1)
            sim = num / denom
            sim = np.where(denom != 0, sim, 0)
            return sim
        elif metric == "jaccard":
            import scipy as sp

            return sp.spatial.distance.jaccard(predicted, target)
        else:
            raise NotImplementedError(f"Metric {metric} is not implemented.")

    def _evaluate(self, x):  # black-box objective function to minimize
        return self.score_function(x, self.target)



import numpy as np
from HDBOBenchmark.algorithms.add_bo_quasi_nt.BOLibkky.preprocessDecomposition import (
    _HyperParam as HyperParam,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt import add_gp_bo


def addbo_optimize(func, bounds, n_iterations, addbo_kwargs=None):
    if addbo_kwargs is None:
        addbo_kwargs = dict()

    decomp_strategy = (
        addbo_kwargs["decomp_strategy"]
        if "decomp_strategy" in addbo_kwargs.keys()
        else "partialLearn"
    )
    n_group = (
        addbo_kwargs["n_group"]
        if "n_group" in addbo_kwargs.keys()
        else int(np.ceil(func.dim / 5))
    )

    hyperParam = HyperParam(func.dim, n_group, True)
    hyperParam.decomp_strategy = decomp_strategy
    best_y, best_x, x_history, y_history, best_y_history = add_gp_bo(
        func, bounds, n_iterations, hyperParam
    )
    result = {
        "best_x": best_x,
        "best_y": best_y,
        "x_history": x_history,
        "y_history": y_history,
        "best_y_history": best_y_history,
    }
    return best_x, best_y, result



import torch
from torch.linalg import cholesky, solve
from .combinedKernelNoise import combined_kernel_noise
from .combinedMeanFunc import combined_mean_func


def add_gp_regression(train_x, train_y, decomposition, hyper_params):
    """
    A python function for perform GP Regression when the model is additive.
    This performs inference in each individual GP.
    Args:
        x: Training input (num, dim)
        y: Traning output
        decomposition: list containing num_groups elements. Each element is a vector containing the coordinates in
            that group.
        hyper_params: Contains the smoothness (sigma_sms), scale (sigma_prs) and noise (noise0, noises) parameters
            for each GP in the additive model.
    Returns:
        mus, k_posts:
        combined_func_h: a function handle for the combined GP.
        func_hs: array with a function handle for each group.
    """
    # Prelims
    num_groups = len(decomposition)

    # Set the hyperparameters for each GP
    # Common Mean Function
    common_mean_func = hyper_params.common_mean_func

    def meanFunc(arg):
        return torch.zeros(arg.shape[0]).to(arg)

    # Mean Functions for each GP
    mean_funcs = [hyper_params.mean_funcs] * num_groups
    # Common Noise parameter
    common_noise = hyper_params.common_noise  # shape()
    # Bandwith parameters
    sigma_sms = hyper_params.sigma_sms  # shape(numGroups)
    # Scale parameters
    sigma_prs = hyper_params.sigma_prs  # shape(numGroups)
    # Noise Parameters
    noises = hyper_params.noises  # shape(numGroups)

    # Construct the Training Kernel Matrix and Invert it
    k0 = combined_kernel_noise(
        train_x, decomposition, sigma_sms, sigma_prs, noises, common_noise
    )
    # To invert this we need to do the cholesky decomposition
    ell = cholesky(k0)
    # compute alpha
    mean_y = combined_mean_func(
        train_x, common_mean_func, mean_funcs, decomposition
    ).reshape(-1, 1)
    train_y = train_y.unsqueeze(-1)  # (num, 1)

    y_ = train_y - mean_y
    alpha = solve(ell.T, (solve(ell, y_)))

    hyper_params.k0 = k0
    hyper_params.ell = ell
    hyper_params.alpha = alpha



import numpy as np
import torch
from HDBOBenchmark.algorithms.turbo import turbo


def turbo_optimize(func, bounds, n_iterations, turbo_kwargs=None):
    if turbo_kwargs is None:
        turbo_kwargs = dict()

    lowB, upB = bounds[:, 0], bounds[:, 1]

    def eval_objective(x):
        """x is assumed to be in [0, 1]^d"""
        lb, ub = torch.tensor(lowB).to(x), torch.tensor(upB).to(x)
        x = lb + (ub - lb) * x
        y = func(x)  # Flip the value for minimization
        return torch.tensor(y).to(x)

    X, Y = turbo(
        eval_func=eval_objective,
        dim=func.dim,
        n_iterations=n_iterations,
        **turbo_kwargs
    )
    X = X.cpu().numpy()
    Y = Y.cpu().numpy()
    best_x, best_y = X[np.argmax(Y)], np.max(Y)
    result = {"best_x": best_x, "best_y": best_y, "x_history": X, "y_history": Y}
    return best_x, best_y, result



import numpy as np
import torch
from .FunctionBase import TestFunction


class Sin(TestFunction):
    def __init__(
        self,
        dim,
        a=1,
        b=1,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(dim, "Sin")

        self.a, self.b = a, b
        self._x_bound = np.array(
            [[-np.pi * 4, np.pi * 4]] * self.dim, dtype=float
        )  # shape(dim, 2)
        self._min_pt = -torch.pi * torch.ones(self.dim) / 2
        self._min_val = -dim
        self.y_scale = 2 * dim
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)
        self.to_torch()

    def _evaluate(self, x):  # black-box objective function to minimize
        return torch.sin(self.a * x + self.b).sum(dim=-1)



from typing import Optional
import numpy as np
import torch
import abc


class TestFunction(object):
    def __init__(self, dim=1, name="TestFunction", x_scaler=None, y_scaler=None):
        self._name = name
        self._dim = dim
        self._x_bound = np.array([[-1, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        self._min_pt = None
        self._min_val = None
        self.y_scale = 100  # For function auto_init_y_scaler

        self._if_torch = False
        self.if_minium = True

    @staticmethod
    def _cal_scaled(a, a_scaler):
        if isinstance(a, np.ndarray) and len(a.shape) == 2:
            return (a + np.tile(a_scaler[0], (a.shape[1], 1)).transpose()) * np.tile(
                a_scaler[1], (a.shape[1], 1)
            ).transpose()
        elif isinstance(a, torch.Tensor) and len(a.shape) == 2:
            return (a + torch.tile(a_scaler[0], (a.shape[1], 1)).T) * torch.tile(
                a_scaler[1], (a.shape[1], 1)
            ).T
        return (a + a_scaler[0]) * a_scaler[1]

    @staticmethod
    def _cal_descaled(a_scaled, a_scaler):
        if isinstance(a_scaled, np.ndarray) and len(a_scaled.shape) == 2:
            return (a_scaled / np.tile(a_scaler[1], (a_scaled.shape[0], 1))) - np.tile(
                a_scaler[0], (a_scaled.shape[0], 1)
            )
        elif isinstance(a_scaled, torch.Tensor) and len(a_scaled.shape) == 2:
            return (
                a_scaled / torch.tile(a_scaler[1], (a_scaled.shape[0], 1))
            ) - torch.tile(a_scaler[0], (a_scaled.shape[0], 1))
        return (a_scaled / a_scaler[1]) - a_scaler[0]

    @property
    def name(self) -> str:
        return self._name

    @property
    def dim(self) -> int:
        return self._dim

    @property
    def bound(self) -> np.ndarray:
        return self._cal_scaled(self._x_bound, self.x_scaler)

    @property
    def min_pt(self) -> Optional[float]:
        if self._min_pt is None:
            return None
        return self._cal_scaled(self._min_pt, self.x_scaler)

    @property
    def min_val(self) -> Optional[float]:
        if self._min_val is None:
            return None
        return self._cal_scaled(self._min_val, self.y_scaler)

    @property
    def if_torch(self) -> bool:
        return self._if_torch

    @abc.abstractmethod
    def _evaluate(self, x) -> Optional[float]:
        return

    def _x_preprocess(self, x):
        """
        Return x: (batch, feature)
        """
        if self._if_torch:
            if not isinstance(x, torch.Tensor):
                x = torch.tensor(x).to(self.x_scaler)
            while x.dim() < 2:
                x = x.unsqueeze(0)
        else:
            if isinstance(x, torch.Tensor):
                x = x.detach().cpu().numpy()
            elif not isinstance(x, np.ndarray):
                x = np.array(x)
            while x.ndim < 2:
                x = x[np.newaxis]
        return x

    def evaluate(self, x) -> dict:
        x = self._x_preprocess(x)
        result = dict()
        result["x"] = x
        y = self(x)
        result["y"] = y
        return result

    def __call__(self, x) -> float:
        x = self._x_preprocess(x)
        x = self._cal_descaled(x, self.x_scaler)
        y = self._evaluate(x)
        y = self._cal_scaled(y, self.y_scaler)
        if not self.if_minium:
            y = -y
        return y

    def init_scaler(
        self,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        self.x_scaler = (
            np.array([np.zeros(self.dim), np.ones(self.dim)], dtype=float)
            if x_scaler is None
            else x_scaler
        )
        self.y_scaler = np.array([0, 1], dtype=float) if y_scaler is None else y_scaler
        if auto_init_x_scaler:
            self.auto_init_x_scaler()
        if auto_init_y_scaler:
            self.auto_init_y_scaler()

    def auto_init_x_scaler(self):
        self.x_scaler[0] = (self._x_bound[:, 1] + self._x_bound[:, 0]) * -0.5
        self.x_scaler[1] = 2 / (self._x_bound[:, 1] - self._x_bound[:, 0])

    def auto_init_y_scaler(self):
        self.y_scaler[0] = -self._min_val
        self.y_scaler[1] = 100 / self.y_scale

    # only if function is in torch
    def to_torch(self, device="cpu"):
        if not self._if_torch:
            self._if_torch = True
            self._min_pt = torch.tensor(self._min_pt, device=device)
            self._x_bound = torch.tensor(self._x_bound, device=device)
            self.x_scaler = torch.tensor(self.x_scaler, device=device)
            self.y_scaler = torch.tensor(self.y_scaler, device=device)

    def switch_maximum(self):
        self.if_minium = False

    def switch_minium(self):
        self.if_minium = True


class BOTorchTestFunction(TestFunction):
    def __init__(
        self,
        func,
        name="BOTorchTestFunction",
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        self.botorch_func = func
        super().__init__(self.botorch_func.dim, name)
        self._x_bound = np.array(self.botorch_func._bounds)  # shape(dim, 2)
        self._min_pt = np.array(self.botorch_func._optimizers[0])
        # self._min_val = self._evaluate(self._min_pt)
        self._min_val = self.botorch_func.optimal_value
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)
        self.to_torch()

    def _evaluate(self, x):  # black-box objective function to minimize
        return self.botorch_func(x)

    def init_scaler(
        self, x_scaler, y_scaler, auto_init_x_scaler=True, auto_init_y_scaler=True
    ):
        super().init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)
        if not isinstance(self.x_scaler, torch.Tensor):
            self.x_scaler = torch.tensor(self.x_scaler)



import os, sys
from pathlib import Path
import pandas as pd

ROOT_DIR = str(Path(os.path.realpath(__file__)).parent.parent)
sys.path.append(ROOT_DIR)
from .design_space import DesignSpace


def parse_int(row: pd.Series):
    name = row.name
    if len(row.range.split(",")) == 2:
        lb, ub = [int(x) for x in row.range.split(",")]
        if lb < ub:
            def_val = row.default
            return {"name": name, "type": "int", "lb": lb, "ub": ub, "default": def_val}
        else:
            values = [int(x) for x in row.range.split(",")]
            def_val = int(row.default)
            return {
                "name": name,
                "type": "cat",
                "categories": values,
                "default": def_val,
            }
    else:
        values = sorted([int(x) for x in row.range.split(",")])
        def_val = int(row.default)
        return {"name": name, "type": "cat", "categories": values, "default": def_val}


# def parse_bool(row: pd.Series):
#     name = row.name  # NOTE: name is index
#     def_val = row['default'].lower() == 'true'
#     return {'name': name, 'type': 'bool', 'default': def_val}


def parse_bool_cat(row: pd.Series):
    name = row.name  # NOTE: name is index
    categories = [0, 1]
    if row.default in ["True", "TRUE"]:
        def_val = 1
    elif row.default in ["False", "FALSE"]:
        def_val = 0
    else:
        RuntimeError(f"default value not recognised {row.default}")
    return {"name": name, "type": "bool", "categories": categories, "default": def_val}


def parse_real(row: pd.Series):
    name = row.name
    lb, ub = [float(x) for x in row.range.split(",")]
    def_val = row.default
    return {"name": name, "type": "num", "lb": lb, "ub": ub, "default": def_val}


def parse_char(row: pd.Series):
    name = row.name
    categories = list(row.range)
    def_val = row.default
    return {"name": name, "type": "cat", "categories": categories, "default": def_val}


def parse_str(row: pd.Series):
    name = row.name
    categories = list(row.range.split(","))
    def_val = row.default
    return {"name": name, "type": "cat", "categories": categories, "default": def_val}


parse_func = {}
parse_func["bool"] = parse_bool_cat
parse_func["char"] = parse_char
parse_func["str"] = parse_str
parse_func["int"] = parse_int
parse_func["real"] = parse_real


def parse_scip_para(file_path: str) -> DesignSpace:
    df = pd.read_csv(file_path, index_col=0)
    para_conf = []
    for i, row in df.iterrows():
        ptype = row.type.lower()
        if ptype not in parse_func.keys():
            print(f"Type {ptype} not supported")
        if ptype in parse_func.keys():
            para_conf.append(parse_func[ptype](row))

    def_val = {para["name"]: para["default"] for para in para_conf}
    design_space = DesignSpace().parse(para_conf)
    return design_space, def_val


if __name__ == "__main__":
    space, default = parse_scip_para("./conf/paras.csv")
    sample = space.sample(1)
    sample = pd.concat([pd.DataFrame([default]), sample], axis=0).reset_index(drop=True)
    print(sample.iloc[:, :5])



import os

import torch

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

from .utils.os import save_pickle
from .utils.logging import logger


def optimize_process(args, method):
    name = args.func.name
    dim = args.func.dim
    logger.info(
        f"func:{name}, dim:{dim}, min:{args.func.min_val}, optimizer:{method.__name__}"
    )

    args.func.switch_maximum()  # for optimizers for maximum
    method_kwargs = dict()
    if "method_kwargs" in dir(args):
        for method_name in ["botorch", "hebo", "turbo", "rembo", "saasbo", "alebo"]:
            if (method_name in method.__name__) and (
                method_name in args.method_kwargs.keys()
            ):
                method_kwargs = args.method_kwargs[method_name]
                break

    best_x, best_y, result_dict = method(
        args.func, args.func.bound, args.numIterations, method_kwargs
    )
    if isinstance(best_x, torch.Tensor):
        best_x = best_x.detach().cpu().numpy()
    if isinstance(best_y, torch.Tensor):
        best_y = best_y.detach().cpu().numpy()
    for k, v in result_dict.items():
        if isinstance(v, torch.Tensor):
            result_dict[k] = v.detach().cpu().numpy()
    best_y, result_dict["best_y"] = -best_y, -best_y
    for k in ["y_history", "best_y_history"]:
        if k in result_dict.keys():
            result_dict[k] = -result_dict[k]

    logger.info(f"Best value found is {best_y:.4f} with point {best_x}.")
    save_cnt = 0
    save_dir = f"./result/{method.__name__}/{name}_{dim}_{save_cnt}.pkl"
    while os.path.exists(save_dir):
        save_cnt += 1
        save_dir = f"./result/{method.__name__}/{name}_{dim}_{save_cnt}.pkl"
    save_pickle(save_dir, result_dict)


def basefunc_tests():
    # get arguments
    from .args.base.ackley import args as a1
    from .args.base.bukin import args as a2
    from .args.base.dropwave import args as a3
    from .args.base.eggholder import args as a4
    from .args.base.griewank import args as a5
    from .args.base.hartmann6d import args as a6
    from .args.base.holdertable import args as a7
    from .args.base.product_sines import args as a8
    from .args.base.rastrigin import args as a9
    from .args.base.rosen_brock import args as a10
    from .args.base.shubert import args as a11
    from .args.base.sphere import args as a12
    from .args.base.trid import args as a13

    arg_list = [a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13]
    # get methods
    from .wrapper import (
        botorch_optimize,
        hebo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
        addbo_optimize,
    )

    method_list = [
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
    ]

    dims = [
        5,
    ]
    for dim in dims:
        for method in method_list:
            for a in arg_list:
                a.func = a.get_func(dim)
                optimize_process(a, method)


def addfunc_tests():
    # get arguments
    from .args.add.HDBOBenchmark import args as a1

    arg_list = [
        a1,
    ]

    # get methods
    from .wrapper import (
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
    )

    method_list = [
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
    ]

    for method in method_list:
        for a in arg_list:
            if isinstance(a.func, list):
                func_list = a.func
                for f in func_list:
                    a.func = f
                    optimize_process(a, method)
                a.func = func_list
            else:
                optimize_process(a, method)


def single_test():
    # get arguments
    from .args.add.HDBOBenchmark import args as a1

    _func = a1.func
    a1.func = [a1.func[0], a1.func[10], a1.func[20]]

    arg_list = [
        a1,
    ]

    # get methods
    from .wrapper import (
        random_optimize,
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
    )

    method_list = [
        random_optimize,
        rembo_optimize,
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        turbo_optimize,
        alebo_optimize,
        saasbo_optimize,
    ]

    for method in method_list:
        for a in arg_list:
            if isinstance(a.func, list):
                func_list = a.func
                for f in func_list:
                    a.func = f
                    optimize_process(a, method)
                a.func = func_list
            else:
                optimize_process(a, method)

    a1.func = _func


def special_test():
    # get arguments
    from .args.realistic.miplib import args as a1

    arg_list = [
        a1,
    ]
    # get methods
    from .wrapper import (
        random_optimize,
        botorch_optimize,
        hebo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
        addbo_optimize,
    )

    method_list = [
        random_optimize,
        botorch_optimize,
        hebo_optimize,
        addbo_optimize,
        alebo_optimize,
        saasbo_optimize,
        rembo_optimize,
        turbo_optimize,
    ]

    for method in method_list:
        for a in arg_list:
            optimize_process(a, method)


if __name__ == "__main__":
    single_test()



import torch
from torch import Tensor
from typing import Callable, List
from torch.linalg import solve
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.combinedKernelNoise import (
    combined_kernel,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.combinedMeanFunc import (
    combined_mean_func,
)


def compute_comb_gp_output(
    x_test: Tensor,
    x_train: Tensor,
    decomposition: List[Tensor],
    ell: Tensor,
    alpha: Tensor,
    bws: Tensor,
    scales: Tensor,
    mean_funcs: List[Callable],
    common_mean_func: Callable,
):
    """
    Returns the predictive mean and variance for the combined GP.
    Args:
        x_test: testing set.
        x_train: training set.
        decomposition: A list with elements that give the coordinates for each group.
        ell: A 'num x num' tensor, which is the kernel matrix of training set.
        alpha:
        bws: A 'numGroups' tensor, which gives bandwidths for each group.
        scales: A 'numGroups' tensor, which gives scales for each group
        mean_funcs: A list containing meanFunction for each group.
        common_mean_func:
    Returns:
        y_mu, y_std, y_k:
    """
    k12 = combined_kernel(x_train, x_test, decomposition, bws, scales)  # shape(num, 1)
    k22 = combined_kernel(x_test, x_test, decomposition, bws, scales)  # shape(1, 1)

    # Compute the outputs
    y_mu = (
        combined_mean_func(x_test, common_mean_func, mean_funcs, decomposition).reshape(
            -1, 1
        )
        + k12.T @ alpha
    )

    v = solve(ell, k12)
    y_k = k22 - v.T @ v
    y_std = torch.sqrt(y_k.diag())
    return y_mu.squeeze(), y_std.squeeze(), y_k
